# FFR Toolbox v1 - Mekong River Basin
# Copyrigth WWF-US, Confluvio Consulting
# 2022-09-20


from ast import And
import arcpy
import numpy as np
import numpy.lib.recfunctions as rfn
import pandas as pd
import os
import sys
import cPickle
import datetime
import logging
import shutil
import stat
import time
import multiprocessing
import math
from collections import defaultdict

# import packages
# from cmath import nan
# issues importing nan in above line
# line below accomplishes the same: https://stackoverflow.com/questions/43428386/importerror-cannot-import-name-inf
math.nan = float('nan')
nan = math.nan


########## create the toolbox ###########
class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "Free_Flowing_Rivers_Analysis"
        self.alias = "Free_Flowing_Rivers_Analysis"

        # List of tool classes associated with this toolbox
        self.tools = [Tool]


########## config functions ##########
# copied an pasted from config.py
class var:
    """
    Class variables to be used throughout the script.
    For description of fields see documentation file at https://doi.org/10.6084/m9.figshare.7688801

    """
    #################
    #  File Paths
    #################
    time_stamp = "time_stamp"
    output_folder = "output_folder"
    config_excel = "config_excel"

    #################
    #  Barrier fields
    #################

    # Storage volume in million cubic meters (MCM)

    STOR_MCM = "STOR_MCM"

    # Discharge factor upstream (if used)
    DFU = "DFU"
    # Discharge factor downstream (if used)
    DFD = "DFD"

    #########################
    #  Benchmark table fields
    #########################

    # Free-flowing river benchmark id
    FFRID = "FFRID"

    # Benchmark source (e.g. study, expert)
    BENCH_SRC = "BENCH_SRC"
    Name_Expert = "Name_Expert"

    #####################
    # River reach fields
    #####################

    REACH_ID = "REACH_ID"

    GOID = "GOID"
    NOID = "NOID"
    NDOID = "NDOID"
    NUOID = "NUOID"

    # Continent ID
    CON_ID = "CON_ID"

    # Continent Name
    CONTINENT = "CONTINENT"

    # Country name
    COUNTRY = "COUNTRY"

    # Basin ID
    BAS_ID = "BAS_ID"

    # Basin Name
    BAS_NAME = "BAS_NAME"

    # Length of the river reach in kilometers
    LENGTH_KM = "LENGTH_KM"

    # Volume of river reach in thousand cubic meters
    VOLUME_TCM = "VOLUME_TCM"

    # Long-term average discharge in cubic meters per second
    DIS_AV_CMS = "DIS_AV_CMS"

    # Discharge river order of river reach (log10 of discharge)
    RIV_ORD = "RIV_ORD"

    # Field indicating if there is a waterfall present (1) or not (0)
    HYFALL = "HYFALL"

    # Upland area in square kilometers
    UPLAND_SKM = "UPLAND_SKM"

    # Percent of river reach catchment covered with floodplains
    FLD = "FLD"

    #######################
    # Backbone river fields
    #######################

    # The fields below define the river reach as part of a 'backbone river'
    # Backbone ID
    BB_ID = "BB_ID"

    # River Name
    BB_NAME = "BB_NAME"

    ISO_NAME = "ISO_NAME"

    # River connected to ocean (1) or not (0)
    BB_OCEAN = "BB_OCEAN"

    # Total length and volume_mcm (sum) of the backbone river
    BB_LEN_KM = "BB_LEN_KM"
    BB_VOL_TCM = "BB_VOL_TCM"

    # Discharge river order of most downstream river reach of backbone river (log10 of discharge)
    BB_DIS_ORD = "BB_DIS_ORD"

    # used as a temporaray field for filter procedure
    CAT_FFR_FILTER = "CAT_FFR_FILTER"

    # Defines if river reach is included in assessment
    # 1 = yes
    # 0 = no
    INC = "INC"

    #

    ###############################
    # Sediment Index Calculations
    ###############################

    # Existing sediment yield of all pixels in reach catchment
    ERO_YLD_TON = "ERO_YLD_TON"

    ###############################
    # Field to create and calculate
    ###############################

    SED_LSS_LKS_OT_NAT = "SED_LSS_LKS_OT_NAT"

    SED_LSS_DMS_ANT = "SED_LSS_DMS_ANT"

    # Total accumulated sediment load taking into account lakes
    SED_NAT = "SED_NAT"
    SED_NAT_UP = "SED_NAT_UP"
    SED_LSS_LKS_IN_NAT = "SED_LSS_LKS_IN_NAT"

    # Total accumulated sediment load taking into account lakes and dams
    SED_ANT = "SED_ANT"
    SED_ANT_UP = "SED_ANT_UP"
    SED_LSS_LKS_OT_ANT = "SED_LSS_LKS_OT_ANT"
    SED_LSS_LKS_IN_ANT = "SED_LSS_LKS_IN_ANT"

    # Total sediment loss
    SED_LSS_TOT = "SED_LSS_TOT"

    # Sediment Trapping Index
    SED = "SED"

    def __init__(self):
        pass


########## tools functions ##########
# define functions from the original .../tools/helper.py module
class tools(object):
    
    fd = var

    def __init__(self):
        pass

    @staticmethod
    def test_64(messages):
        tools.prt(messages,"Running against: {}".format(sys.version))
        if sys.maxsize > 2 ** 32:
            tools.prt(messages,"Running python 64 bit")
        else:
            tools.prt(messages,"Running python 32 bit")

    @staticmethod
    def get_stamp():
        """
        Creates a time stamp in a specific format
        :return: timestamp
        """
        stamp1 = int(
            datetime.datetime.fromtimestamp(time.time()).strftime('%y%m%d'))
        stamp21 = int(datetime.datetime.fromtimestamp(time.time()).strftime('%H'))
        stamp22 = int(datetime.datetime.fromtimestamp(time.time()).strftime('%M'))
        stamp23 = int(datetime.datetime.fromtimestamp(time.time()).strftime('%S'))
        stamp3 = int(datetime.datetime.fromtimestamp(time.time()).strftime('%f'))

        stamp = str(stamp1) + "_" + str(stamp21).zfill(2) + str(stamp22).zfill(
            2) + str(stamp23).zfill(2) + "_" + str(
            stamp3)

        return stamp

    @staticmethod
    def check_fields(table, flds, messages):
        """

        Check if multiple fields exist

        :param table: Table to check
        :param flds: Fields to check

        """
        fld_list = arcpy.ListFields(table)
        x = [a.name for a in fld_list]

        count_wrong = 0
        for fld in flds:
            if fld not in x:
                arcpy.AddWarning("Field {} does not exist".format(fld))

                count_wrong += 1

        if count_wrong > 0:
            arcpy.AddError ("There were {} fields missing from the table. "
                "Please add the fields to the table before "
                "proceeding".format(count_wrong))
            sys.exit()

    @staticmethod
    def check_field(table, fld):
        """
        Checks if one field exists

        :param table: Table to check
        :param fld: Field to check

        """
        fld_list = arcpy.ListFields(table)
        x = [a.name for a in fld_list]

        if fld not in x:
            return True
        else:
            return False

    @staticmethod
    def check_esri_item(path_to_item, messages):
        """
        Checks if an ESRI item exists

        :param path_to_item:
        :return:
        """

        if arcpy.Exists(path_to_item):
            pass
        else:
            tools.prt(messages,"%s does not exist" % path_to_item)
            sys.exit(-99)

    @staticmethod
    def create_path(path, messages):
        """
        Checks if a path exists

        :param path: path to check
        :return: nada
        """
        try:
            os.makedirs(path)
        except Exception as ErrorDesc:
            arcpy.AddError ("Error creating path. " + str(ErrorDesc))

    @staticmethod
    def copytree(src, dst, messages, symlinks=False, ignore=None):
        """
        Copies an entire directory, in this case the code, into the output
        directory

        see https://stackoverflow.com/a/22331852/344647

        :param src: source directory
        :param dst: destination directory
        :param symlinks:
        :param ignore:
        :return:
        """
        if not os.path.exists(dst):
            os.makedirs(dst)
            shutil.copystat(src, dst)
        lst = os.listdir(src)
        if ignore:
            excl = ignore(src, lst)
            lst = [x for x in lst if x not in excl]
        for item in lst:
            s = os.path.join(src, item)
            d = os.path.join(dst, item)
            if symlinks and os.path.islink(s):
                if os.path.lexists(d):
                    os.remove(d)
                os.symlink(os.readlink(s), d)
                try:
                    st = os.lstat(s)
                    mode = stat.S_IMODE(st.st_mode)
                    os.lchmod(d, mode)
                except Exception as ErrorDesc:
                    arcpy.AddError ("Error copying tree. " + str(ErrorDesc))

            elif os.path.isdir(s):
                try:
                    tools.copytree(s, d, symlinks, ignore)
                except Exception as e:
                    arcpy.AddError (str(e) + "Perhaps the EXCEL file is still open?")
            else:
                try:
                    shutil.copy2(s, d)
                except Exception as e:
                    arcpy.AddError (str(e) + "Perhaps the EXCEL file is still open?")

    @staticmethod
    def add_fields(array, desc):
        """
        Adds fields to a numpy array

        see: https://stackoverflow.com/a/1201821/344647

        :param array:
        :param desc:
        :return:
        """
        if array.dtype.fields is None:
            arcpy.AddMessage("A must be a structured numpy array")

        b = np.empty(array.shape, dtype=array.dtype.descr + desc)
        for name in array.dtype.names:
            b[name] = array[name]
        return b

    @staticmethod
    def delete_field(esri_table, list_of_field_names, messages):
        """

        Deletes a field

        :param esri_table: esri feature class or table
        :param list_of_field_names: field name to delete
        :return:
        """

        for f in list_of_field_names:
            if tools.check_field(table=esri_table, fld=f):
                try:
                    arcpy.DeleteField_management(esri_table, f)
                    tools.prt(messages,"Deleted %s" % f)

                except Exception as ErrorDesc:
                    arcpy.AddError ("Error deleting a field " + f + ". " + arcpy.AddError(str(ErrorDesc)))

    @staticmethod
    def add_index(lyr, field_name, messages):
        try:
            if not tools.index_exists(lyr, field_name):
                arcpy.AddIndex_management(lyr, field_name, field_name, "UNIQUE", "ASCENDING")
        except Exception as error:
            tools.prt(messages,"Adding index {} not successful. {}".format(field_name, str(error)))

    @staticmethod
    def index_exists(tablename, indexname):
        if not arcpy.Exists(tablename):
            return False

        tabledescription = arcpy.Describe(tablename)

        for iIndex in tabledescription.indexes:
            if iIndex.Name == indexname:
                return True
        return False

    @staticmethod
    def update_stream_routing_index(streams):
        """
        Function to recalculate the Network OIDs of a stream network, if the
        stream network was extracted from the global network. In this case the
        NOIDs used for the routing are out of sync with the object ids and
        with the numpy indices. The object with the network id 1 must be the
        first object in the list etc.

        args:
            streams(numpy array): numpy array representing the stream network
        return:
            Updates and returns the stream network with new network ids
        """

        for n in ["GOID", "NOID", "NDOID", "NUOID"]:
            if not n in streams.dtype.names:
                raise Exception("Field {} does not exist".format(n))

        # river network must be global or extracted from global. NOID NDOID and
        # NUOID must not be sorted
        if streams["NOID"][0] != streams["GOID"][0]:
            raise Exception("It seems like you are attempting to resort "
                            "a network that has already been sorted. "
                            "This would be confusing, please use an extract of "
                            "the original network")

        # Create Routing Dictionaries and fill

        oidDict = {}
        upsDict = defaultdict(str)
        downDict = defaultdict(long)

        i = 1
        for myrow in streams:
            oidDict[int(myrow[tools.fd.GOID])] = i
            i += 1

        i = 1
        for myrow in streams:
            dnOldOID = myrow[tools.fd.NDOID]
            newOid = oidDict.get(int(dnOldOID), -1)

            if newOid != -1:
                # Write OID of next downstream reach
                downDict[i] = newOid
                # Write OID of next upstream reach
                exValue = upsDict.get(int(newOid), -99)

                if exValue == -99:
                    upsDict[int(newOid)] = i
                else:
                    newV = str(exValue) + '_' + str(i)
                    upsDict[int(newOid)] = newV
            i += 1

        # Writing index values back to numpy
        i = 1
        for myrow in streams:
            myrow[tools.fd.NOID] = i
            myrow[tools.fd.NDOID] = downDict[i]
            myrow[tools.fd.NUOID] = upsDict[i]

            i = i + 1

        return streams

    @staticmethod
    def delete_path(path, messages):
        try:
            if os.path.exists(path):
                shutil.rmtree(path)

        except Exception as err:
            tools.prt(messages,"Cannot remove folder")
            tools.prt(messages,str(err))

    @staticmethod
    def create_path(path, messages):
        try:
            if not os.path.exists(path):
                os.makedirs(path)
        except Exception as err:
            arcpy.AddError ("Cannot make directory. " + str(err))

    @staticmethod
    def copy_between(to_join_fc, to_join_field, IntoJoinField, FromJoinFC,
                    FromJoinField, FromValueField, messages, over_mode=True,
                    over_value=0):
        """
        Copies data between tables
        """

        tools.prt(messages,"Updating tables")

        from_valueDict = defaultdict(float)

        to_flds = [to_join_field, IntoJoinField]

        # Loading the values from FROM table
        from_flds = [FromJoinField, FromValueField]
        with arcpy.da.SearchCursor(FromJoinFC, from_flds) as cursor:
            for myrow in cursor:
                from_valueDict[myrow[0]] = myrow[1]

        # Writing values to TO table
        with arcpy.da.UpdateCursor(to_join_fc, to_flds) as cursor:
            for row1 in cursor:
                ToJoinFieldValue = row1[0]
                newValue = from_valueDict.get(ToJoinFieldValue)

                if newValue is None:
                    if over_mode is True:
                        # This overwrites the values in the TO table
                        # with zero values for IDs not present in FROM table
                        # Default for updating DOR DOF results

                        # If you just want to update certain values from one table
                        # to the other and leave all other rows untouched,
                        # then use the "update" over_mode

                        row1[1] = over_value
                else:
                    row1[1] = newValue

                cursor.updateRow(row1)

    @staticmethod
    def get_writer(base_dir, stamp):
        # Setup a Excel Writer target file
        excel_file_output = os.path.join(base_dir, "results_" + stamp + ".xls")
        return pd.ExcelWriter(excel_file_output), excel_file_output

    @staticmethod
    def export_excel(df, name, writer, id=False, startrow=0,
                    header=False, first=False):
        try:
            startrow = len(writer.sheets[name].rows)
        except BaseException:
            startrow = 0

        if startrow == 0:
            header = True
        else:
            header = False

        if first:
            header = False

        df.to_excel(
            writer,
            name,
            index=False,
            encoding='utf-8',
            startrow=startrow,
            header=header)
        writer.save()

    @staticmethod
    def create_results_sheet(writer):
        result_list = ["Stamp",
                    "Sce_name",
                    "I_1",
                    "I_2",
                    "I_3",
                    "I_4",
                    "I_5",
                    "I_6",
                    "W_1",
                    "W_2",
                    "W_3",
                    "W_4",
                    "W_5",
                    "W_6",
                    "Threshold",
                    "Flood_weight",
                    "Export",
                    "Filter_Threshold",
                    "Sce_name",
                    "Count_reaches",
                    "Count_reaches_affected",
                    "Perc_reaches_affected",
                    "Mean_csi",
                    "count_nff",
                    "perc_nff",
                    "Bench_match"]
        results_pd = pd.DataFrame(result_list).T
        results_pd.drop(results_pd.index[[0]])

        tools.export_excel(results_pd, "Global_stats", writer, first=True)

        dom_list = ["Stamp",
                    "Sce_name",
                    "Pressure",
                    "Number of reaches"]
        dom_pd = pd.DataFrame(dom_list).T
        dom_pd.drop(dom_pd.index[[0]])
        tools.export_excel(dom_pd, "Global_dom", writer, first=True)

        dom_bench_list = ["Stamp",
                        "Sce_name",
                        "FFR_ID",
                        "Rivername",
                        "Pressure",
                        "Number of reaches",
                        "Source"
                        ]
        dom_bench_pd = pd.DataFrame(dom_bench_list).T
        dom_bench_pd.drop(dom_bench_pd.index[[0]])
        tools.export_excel(dom_bench_pd, "Bench_dom", writer, first=True)

    @staticmethod
    def load_parameters(path, messages):

        # create two dictionary objects
        start_dict = defaultdict()
        value_dict = defaultdict()

        # Check spreadsheet

        # Does it exist? Opening Excel sheet
        try:
            # open the excel file using pandas
            xls = pd.ExcelFile(path)
            sheet_names = xls.sheet_names
        except Exception as err:
            # create error if the excel sheet cannot be opened
            arcpy.AddError ("Error opening file. Does it exist? {}").format(str(err))
            sys.exit()

        # Checking the sheets.
        # Does it have the start sheet?

        # check if the sheet named START is in the list of sheet names
        if "START" not in sheet_names:
            # if no start sheet exists, create error
            messages.addMessage("\r\n" + "Sheet START does not exist")
            sys.exit()

        # check that the excel file has atleast three sheets
        if len(sheet_names) < 3:
            arcpy.AddError ("There must be at least three sheets in the workbook.")
            arcpy.AddError ("1. The sheet 'START'")
            arcpy.AddError ("2. The sheet settings sheet")
            arcpy.AddError ("3. The sheet scenarios sheet'")


        try:
            # parse the start sheet
            settings_sheet = xls.parse("START")

        except Exception as e:
             # if no start sheet exists, create error
            arcpy.AddError ("Does sheet 'START' exist? {}").format(str(e))
            

        # create two variables to store values from the key column and
        # value cloumn of the start sheet
        var1 = settings_sheet['Key']
        var2 = settings_sheet['Value']

        # create a counter starting at 0
        ind = 0
        # for each entry in var 1
        for fname in var1:
            # the value associated with fname in start dict is assigned as
            # the corresponding value in var2
            start_dict[fname] = var2[ind]
            # increase the counter by 1
            ind += 1

        # iterates throught the list of strings and checks if they are in the
        # 'keys' of the start dic. If not an error is raised
        for sname in ["settings_sheet", "scenarios_sheet", "run_dof", "run_dor", "run_sed", "run_csi"]:
            if sname not in start_dict.keys():
                arcpy.AddError("\r\n" + "Start variable not found: {}".format(sname))

        ### Start sheet parameters ###

        # assigns the settings and scenarios sheet values in the start_dict to variable names   
        set_sheet = start_dict["settings_sheet"]
        sce_sheet = start_dict["scenarios_sheet"]

        sequence = {"run_dof": start_dict["run_dof"],
                    "run_dor": start_dict["run_dor"],
                    "run_sed": start_dict["run_sed"],
                    "run_csi": start_dict["run_csi"]}

        ### settings sheet parameters ###
        # Load project values
        try:
            # parses the settings sheet
            settings_sheet = xls.parse(set_sheet)

        except Exception as e:
            # creates an error if the settings sheet cannot be parsed
            arcpy.AddError (str(e) + "." + " Does sheet {} exist?".format(set_sheet))

        # create two variables to store values from the key column and
        # value cloumn of the settings sheet
        var1 = settings_sheet['Key']
        var2 = settings_sheet['Value']

        # create a counter startng at 0
        ind = 0
        # for each entry in var1
        for fname in var1:
            # the value associated with fname in value_dict is assigned the
            # corresponding value in var2
            value_dict[fname] = var2[ind]
            # increase the counter by 1
            ind += 1

        ### Load scenario list ###
        # create an empty list
        sce_list = []

        try:
            # parse the scenario sheet
            sce_sheet = xls.parse(sce_sheet)
        except Exception as e:
            # if the scenario sheet cannot be parsed, create an error
            arcpy.AddError (str(e))
            arcpy.AddError ("Does sheet {} exist?".format(sce_sheet))

        # Create a list of header names that should be in the scenario sheet
        header = ["scenario_name",
                "dof_indicator",
                "dor_indicator",
                "sed_indicator",
                "use_indicator",
                "rdd_indicator",
                "urb_indicator",
                "dof_weight",
                "dor_weight",
                "sed_weight",
                "use_weight",
                "rdd_weight",
                "urb_weight",
                "csi_thresh",
                "fld_damp",
                "filter_thresh",
                "to_process",
                "to_export"]

        # iterate through the list of headers
        for h in header:
            # if a header is not in the list of headers
            if h not in sce_sheet.columns:
                messages.addMessage("{} not found".format(h))

        # create and empty set
        field_set = set()

        # iterate through each row of the scenarios sheet
        for index, row in sce_sheet.iterrows():
            scename = row[0]
            # crate list of indicator fields
            list_of_fields = [row[1], row[2], row[3], row[4], row[5], row[6]]
            # create list of weight values
            list_of_weights = [row[7], row[8], row[9], row[10], row[11], row[12]]
            # assign other values in the scenario sheet to variables
            csi_thresh = row[13]
            flood_damp = row[14]
            filter_thresh = row[15]
            process = row[16]
            export = row[17]

            # combines the field set and the list of fields
            field_set |= set(list_of_fields)
            
            # appends the scenario values for the row being iterated to the sce_list
            sce_list.append([scename, list_of_fields, list_of_weights,
                            csi_thresh, flood_damp, filter_thresh, process,
                            export])

        return sequence, value_dict, sce_list, list(field_set)

    @staticmethod
    def load_stream_array(stream_feature_class, stream_fields, messages, use_npy=0, fname="", ):
        """
        Loading input stream feature class (river network) to numpy array.
        Optionally, the stream feature class can bes saved as a numpy memory
        object and loaded instead, which can be much more efficient (depending
        on extent)

        :param stream_feature_class:
        :param stream_fields:
        :param use_npy:
        :param fname:
        :param messages:  
        :return:
        """

        def save_arr(arr, loc):
            try:
                np.save(loc, arr)
            except BaseException:
                tools.prt(messages,"Unexpected error:", sys.exc_info()[0])
                raise

        def load_arr(loc):
            try:
                return np.load(loc)
            except BaseException:
                tools.prt(messages,"Unexpected error:", sys.exc_info()[0])
                raise

        if use_npy == 1:
            if os.path.isfile(fname):
                tools.prt(messages,"Loading stream array from npy object")
                arr = load_arr(fname)
            else:
                tools.prt(messages,"Loading stream array from feature class, then save as "
                    "array")
                arr = arcpy.da.TableToNumPyArray(stream_feature_class,
                                                stream_fields)
                tools.prt(messages,"Saving stream array to npy object")
                save_arr(arr, fname)
        else:
            tools.prt(messages,"Loading stream array from feature class")
            arr = arcpy.da.TableToNumPyArray(stream_feature_class, stream_fields)

        # Test if Global network or extracted network
        # If extracted, the routing index must be updated
        tools.update_stream_routing_index(arr)
        return arr

    @staticmethod
    def create_gdb(gdb_folder, gdb_name):
        """
        Creates a path and a geodatabase

        :param gdb_folder:
        :param gdb_name:
        :return:
        """

        if not os.path.exists(gdb_folder):
            os.makedirs(gdb_folder)

        gdb_file_name = gdb_name + ".gdb"
        gdb_full_path = gdb_folder + "\\" + gdb_file_name

        arcpy.CreateFileGDB_management(gdb_folder, gdb_file_name)

        return gdb_full_path, gdb_file_name

    @staticmethod
    def create_mxd(template_mxd, new_mxd_name, template_data_list, stamp, output_folder, messages):
        """

        :param template_mxd: Template database used to copy and save with replaced paths
        :param new_mxd_name:
        :param template_data_list: [ [layer name, layer datasource, new file name, definition ],...]  
        :param stamp:
        :param output_folder:
        :return:
        """


        # provide access to the input template_mxd
        mxd = arcpy.mapping.MapDocument(template_mxd)
        
        # iterate through each dataframe in the template_mxd
        for df in arcpy.mapping.ListDataFrames(mxd, ''):
            # list the layers in the template_mxd
            mxd_lyrs = arcpy.mapping.ListLayers(mxd, '', df)

            # iterate through each layer in the template_mxd
            for lyr in mxd_lyrs:
                
                # check the long name of the lyr

                # check if the layer being iterated has the datasource property
                if lyr.supports("DATASOURCE"):
                    # iterate through each list in the template_data_list list of lists 
                    for i in range(len(template_data_list)):
                        # assign the entries in list being iterated to variable names
                        long_name = template_data_list[i][0]
                        replace_data_source = template_data_list[i][1]
                        new_file_name = template_data_list[i][2]
                        def_query = template_data_list[i][3]
                        # if the longName of the current layer being iterated matches the longName in the template_data_list being iterated
                        if long_name == lyr.longName:

                            # the layers data source is replaced using the corresponding data source and file name in the template_data_list
                            lyr.replaceDataSource(
                                replace_data_source,
                                "FILEGDB_WORKSPACE",
                                 new_file_name )
                            # the defintion query is set using the corresponding definition query in the template_data_list
                            lyr.definitionQuery = def_query

        # get the extension of the template_mxd
        front, extension = os.path.splitext(template_mxd)
        # create the path for the output results mxd
        new_path_to_mxd = os.path.join(output_folder, new_mxd_name + "_" + stamp + extension)
        # save the results mxd using the file path created above
        mxd.saveACopy(new_path_to_mxd)

        return new_path_to_mxd

    @staticmethod
    def export_joined(output_geodatabase_path, output_table_name, table_to_join,
                    join_table, messages):
        """
        Joins the temporary output to the join_fc_lyr feature class and writes it to geodatabase
        """

        arcpy.env.qualifiedFieldNames = False  # to maintain proper field names

        tools.prt(messages,"Adding indices")

        tools.check_esri_item(join_table, messages)
        tools.add_index(lyr=join_table, field_name="GOID", messages = messages)
        tools.add_index(lyr=table_to_join, field_name="GOID", messages = messages)

        csi_fc = os.path.join(output_geodatabase_path,output_table_name)

        # Copy the join feature class to the output destination
        tools.prt(messages,"Joining features")
        try:
            join_name = "join_fc_lyr" + str(output_table_name)
            join_fc_lyr = arcpy.MakeFeatureLayer_management(join_table, join_name)
            joined = arcpy.AddJoin_management(
                join_fc_lyr, "GOID", table_to_join, "GOID", "KEEP_ALL")

            tools.prt(messages,"Exporting features")

            arcpy.CopyFeatures_management(joined, csi_fc)

            tools.delete_field(esri_table=csi_fc,
                        list_of_field_names=["OBJECTID_1", "GOID_1"],
                        messages = messages)

        except Exception as err:
            tools.prt(messages,"Error joining or exporting feature class. {}".format(str(err)))
            sys.exit(0)

        return csi_fc

    @staticmethod
    def get_ffr_field_names(name):
        ffr_stat1 = str(name) + "_FF1"
        ffr_stat2 = str(name) + "_FF2"
        ffr_dis = str(name) + "_FFID"

        new_fields = [(ffr_stat1, 'i4'), (ffr_stat2, 'i4'), (ffr_dis, 'u4')]

        return ffr_stat1, ffr_stat2, ffr_dis, new_fields

    @staticmethod
    def get_csi_field_names(name):
        csi_name = str(name)
        dom_name = str(name) + "_D"
        ff_thresh = str(name) + "_FF"

        new_fields = [(csi_name, 'f8'), (dom_name, '|S3'), (ff_thresh, 'i4')]

        return csi_name, dom_name, ff_thresh, new_fields

    @staticmethod
    def setup_logging(output_folder, messages):
        logfile_path = os.path.join(output_folder, "log.txt")

        logging.basicConfig(
            level=logging.DEBUG,
            filename=logfile_path,
            filemode="a+",
            format="%(asctime)-15s %(message)s")
        return logging

    @staticmethod
    def pd_to_np(nparay, list_of_fields="all"):
        """

        # Turn panda to numpy array
        # https://my.usgs.gov/confluence/display/cdi/pandas.DataFrame+to+ArcGIS+Table

        :param nparay:
        :param list_of_fields:
        :return:
        """
        panda_df = pd.DataFrame(nparay)

        if list_of_fields != "all":
            panda_df = panda_df[list_of_fields]

        # Turn panda to numpy
        # https://my.usgs.gov/confluence/display/cdi/pandas.DataFrame+to+ArcGIS+Table
        x = np.array(np.rec.fromrecords(panda_df.values))
        names = panda_df.dtypes.index.tolist()
        x.dtype.names = tuple(names)

        return x

    @staticmethod
    def remove_csi_traces(fc, name):
        """
        Removes all traces of MIN_, MAX_ etc from dissolve operation output
        For example, the fields generated for each scenario will have the name
        "CSI001", "CSI001_D", "CSI001_FF" etc, if the scenario name (as defined in
        the Excel is "CSI001". To make mapping easier in the GIS, we remove the
        scenario names with "CSI". The results afterwards will be
        "CSI", "CSI_D", "CSI_FF"

        :param fc: feature class
        :param name:
        :return:
        """

        rename_list = [name]

        fieldList = arcpy.ListFields(fc)  # get a list of fields for each feature class
        for field in fieldList:  # loop through each field
            for ren in rename_list:
                ren = ren.replace(" ", "") 
		        
                if field.name.startswith(ren):
                    fn = field.name.replace(ren, 'CSI')
                    try:
                        arcpy.AlterField_management(fc, field.name, fn, fn)
                    except:
                        arcpy.AddError("Error altering the field name for field: {}. fn = {}. Please run the tool again". format(field.name, fn))
                        sys.exit()

			    
                        
    @staticmethod
    def save_as_cpickle(pickle_object, folder, name, file_extension):
        outfile = os.path.join(folder, str(name) + file_extension)
        with open(outfile, 'wb') as fp:
            cPickle.dump(pickle_object, fp)

    @staticmethod
    def update_dam_routing_index(dams, arr):
        """
        Function to recalculate the Global Network OIDs to match with the
        reduced network data set

        args:
            dams(numpy array): numpy array representing a list of dams
            arr(numpy array): numpy array representing the stream network
        return:
            Updates and returns the dam numpy array with new network ids
        """
        for dam in dams:
            oldGOID = dam[tools.fd.GOID]
            find = arr[(arr[tools.fd.GOID]) == oldGOID]
            new_goid = find[tools.fd.NOID]
            dam[tools.fd.GOID] = new_goid

    @staticmethod
    def prt(messages, txt):
        logging.info(txt)
        messages.AddMessage(txt)

########## stats functions ##########
# define functions from the original .../stats/sensitivity.py module
class sens(object):

    def __init__(self):
        pass

    @staticmethod
    def pst_csi_calculations(dir, messages):
        """
        Conducts sensitivity analysis of CSI values

        :param dir: directory where pickles have been saved
        :return: saves a geodatabase table with CSI statistics
        """
        gdb_full_path, gdb_file_name = tools.create_gdb(dir, "stats_csi")

        # create path to the stats_CSI_gdb

        tools.prt(messages,"Processing sensitivity analysis")
        
        mc_results = []

        x = 0
        for filename in os.listdir(dir):

            # NAME.csi files represent scenario run with CSI values of each river reach
            if filename.endswith(".csi"):
                x = x + 1
                tools.prt(messages,filename)
                fil = os.path.join(dir, filename)
                # Load the pickle options back into model
                # https://stackoverflow.com/a/899199/344647
                with open(fil, 'rb') as fp:
                    result_slice = cPickle.load(fp)

                    if x == 1:
                        previous_sum = pd.Series(np.zeros_like(result_slice))
                        previous_diffsquared = pd.Series(
                            np.zeros_like(result_slice))
                        previous_min = pd.Series(result_slice)
                        previous_max = pd.Series(result_slice)

                    sum = (result_slice + previous_sum)
                    mean = sum / float(x)

                    diffsquared = (result_slice - mean) ** 2
                    sum_diffsquared = diffsquared + previous_diffsquared
                    stddev = np.sqrt(sum_diffsquared / float(x))

                    min = np.minimum(result_slice, previous_min)
                    max = np.maximum(result_slice, previous_max)

                previous_sum = pd.Series(sum)
                previous_min = pd.Series(min)
                previous_max = pd.Series(max)
                previous_diffsquared = pd.Series(sum_diffsquared)

            else:
                pass

        mc_results.append(mean)
        mc_results.append(previous_min)
        mc_results.append(previous_max)
        mc_results.append(previous_max - previous_min)
        mc_results.append(stddev)

        concatentated = pd.concat(mc_results, axis=1)

        tools.prt(messages,"Percentile Stats")
        df = pd.DataFrame(concatentated)
        df.columns = ['CSI_AVG', 'CSI_MIN', 'CSI_MAX', 'CSI_RNG', 'CSI_STD']

        # Turn panda to numpy
        # https://my.usgs.gov/confluence/display/cdi/pandas.DataFrame+to+ArcGIS+Table
        x = np.array(np.rec.fromrecords(df.values))
        names = df.dtypes.index.tolist()
        x.dtype.names = tuple(names)
        arcpy.da.NumPyArrayToTable(x, gdb_full_path)

# define functions from the originial .../stats/global_stats.py module
class gstat(object):

    def __init__(self):
        pass

    @staticmethod
    def post_stats_global_single(stream_array, csi_name, csi_threshold):
        """
        Calculate a series of global statistics and write into Excel sheet as row

        :param stream_array:
        :param csi_name:
        :param csi_threshold:
        :return:
        """
        global threshold
        panda0 = pd.DataFrame(stream_array)
        panda_df = panda0[panda0[benchmark.fd.INC] == 1]

        glo = []
        glo.append(["sce_name", "count_reaches", "count_reaches_affected",
                    "perc_reaches_affected", "mean_csi", "count_nff", "perc_nff"])
        headers = glo.pop(0)

        return gstat.aggregate_global_stats(panda_df, csi_name, csi_threshold)

    @staticmethod
    def aggregate_global_stats(df, sce, threshold):
        """
        Calculate global statistics

        :param df:
        :param sce:
        :param threshold:
        :return:
        """

        # "Global count reaches"
        reach_count = df[sce].count()

        # "Global count reaches impacted (CSI < 100)"
        glb = df[df[sce] < 99.99999999999999]
        imp_count = glb[sce].count()

        # "Percent of global impacted (CSI < 100)"
        pct_global_imp = round((100 * (imp_count / float(reach_count))), 1)

        # "Global Average (CSI < 100)"
        glb = df[df[sce] < 99.99999999999999]
        imp_mean = round(glb[sce].mean(), 1)

        # "Count affected reaches (CSI < thres)"
        glb2 = df[df[sce] < threshold]
        nff = glb2[sce].count()

        # "Percent NNF reaches (CSI < thres)"
        pct_nff = round((100 * (nff / float(reach_count))), 1)

        return [
            sce,
            reach_count,
            imp_count,
            pct_global_imp,
            imp_mean,
            nff,
            pct_nff]

    @staticmethod
    def export_global_stats_results_to_excel(name_sheet, result_list, writer):
        """
        Now export global stats with benchmarking results

        :param name_sheet:
        :param result_list:
        :param writer:
        :return:
        """

        # Converting list to panda data frame
        # https://stackoverflow.com/a/53944118/344647

        results_pd = pd.DataFrame(result_list).T
        tools.export_excel(df=results_pd,
                        name=name_sheet,
                        writer=writer)

    @staticmethod
    def write_bench_results_to_global_results(bench_val, result_list):
        result_list += bench_val
        return result_list

# define functions from the original .../stats/benchmarking.py module
class benchmark(object):
    
    fd = var

    def __init__(self):
        pass

    @staticmethod
    def post_stats_bench_single(
            stream_array_mod,
            scenario_name,
            bench_fc,
            csi_threshold, 
            messages):

        """
        Calculates benchmarking statistics

        :param stream_array_mod:
        :param scenario_name:
        :param bench_fc:
        :param csi_threshold:
        :return:
        """

        bench = benchmark._loadBenchTable(bench_fc)
        bench = bench.set_index([benchmark.fd.GOID])

        # Now preparing benchmark tables
        panda_df_for_bench = pd.DataFrame(stream_array_mod)
        panda_df_for_bench = panda_df_for_bench.set_index([benchmark.fd.GOID])

        join_bench = pd.merge(
            panda_df_for_bench,
            bench,
            left_index=True,
            right_index=True)

        dom_field_name = scenario_name + str("_D")
        dom_bench = benchmark.calculate_dominance_bench_rivers(
            join_bench, dom_field_name, scenario_name, csi_threshold, messages)

        # Benchmarking stats requires two steps.
        number_matching_bench_rivers = benchmark.benchmarking_rivers(join_bench,
                                                        scenario_name,
                                                        csi_threshold,
                                                        messages)

        return number_matching_bench_rivers, dom_bench

    @staticmethod
    def calculate_dominance_bench_rivers(join, domField, FieldName, thres, messages):
        """
        Function to determine the DOM index for rivers that failed benchmarking

        :param join:
        :param domField:
        :param FieldName:
        :param thres:
        :return:
        """
        tools.prt(messages,
            "DOM: Calculating scenario {0} with threshold {1}".format(
                FieldName,
                thres))
        join.loc[:, "SCE_NAME"] = FieldName
        join.loc[:, "NUM"] = 1
        sel = join[join[FieldName] < thres]


        if len(sel) == 0:
            tools.prt(messages,'All benchmark rivers for scenario {0} were identified as free-flowing using CSI threshold {1}'.format(FieldName, thres))
            
            dom = pd.DataFrame({
            'NUM': ['All benchmark rivers identified as free-flowing'],
            'SCE_NAME': [FieldName],
            'BENCH_SRC': ['All benchmark rivers identified as free-flowing'],
            "Name_Expert": ['All benchmark rivers identified as free-flowing'],
            'FFRID': ['All benchmark rivers identified as free-flowing'],
            'Pressure': ['No pressure identified']})

            return dom 

        fun = {
            'NUM': np.sum,
            'SCE_NAME': 'first',
            'BENCH_SRC': 'first',
            "Name_Expert": 'first'}
        dom = sel.groupby(["FFRID", domField], as_index=False).agg(fun)
        dom.rename(columns={domField: 'Pressure'}, inplace=True)
        return dom

    @staticmethod
    def benchmarking_rivers(join, sce, threshold, messages):
        """
        Calculates number of matching rivers

        :param join:
        :param sce:
        :param threshold:
        :return:
        """
        join.loc[:, "NUM"] = 1  # Helps in creating a count field

        # First I am creating statistics for each benchmark river (FFRID)
        # sce: np.min is getting the lowest CSI value for the river
        # Will be compared later with threshold in function
        f = {
            benchmark.fd.BB_ID: 'first',
            benchmark.fd.BAS_NAME: 'first',
            benchmark.fd.BB_NAME: 'first',
            benchmark.fd.ISO_NAME: 'first',
            benchmark.fd.LENGTH_KM: np.sum,
            benchmark.fd.VOLUME_TCM: np.sum,
            benchmark.fd.RIV_ORD: np.min,
            sce: np.min,
            benchmark.fd.FFRID: 'first',
            benchmark.fd.BENCH_SRC: 'first',
            benchmark.fd.Name_Expert: 'first'}

        one = join.groupby(benchmark.fd.FFRID).agg(f)
        # tools.prt(messages,"one")
        # tools.prt(messages,one)

        # Then calculating if above threshold

        # Assigning value of 1 if entire river is above threshold, or 0 if not
        one.loc[:, "MATCH"] = one[sce].apply(benchmark._getLevelCode, var1=threshold)

        one.loc[:, "SCE_NAME"] = sce
        one.loc[:, "NUM"] = 1

        # Aggregating to number of free-flowing rivers / non-free flowing rivers
        fun = {'NUM': np.sum, "SCE_NAME": 'first'}
        # tools.prt(messages,"fun")
        # tools.prt(messages,fun)


        two = one.groupby(["MATCH"], as_index=False).agg(fun)
        # tools.prt(messages,"two_1")
        # tools.prt(messages,two)

        two.rename(columns={"NUM": 'COUNT'}, inplace=True)
        # tools.prt(messages,"two_2")
        # tools.prt(messages,two)

        # Selecting only rows where Match = 1 (free-flowing rivers)
        two = two[two["MATCH"] == 1]

        # tools.prt(messages,"len_two")
        # tools.prt(messages,len(two))
        # tools.prt(messages,"two_3")
        # tools.prt(messages,two)
        
        # check if there are no matches
        if len(two) == 0:
            number_matching_rivers = 0            
            return number_matching_rivers

        else:
            number_matching_rivers = two.iloc[0]["COUNT"]
            return number_matching_rivers

    @staticmethod
    def _load_streams(table, fields="*"):
        arr = arcpy.da.TableToNumPyArray(table, fields, null_value=0)
        return arr

    @staticmethod
    def _getLevelCode(x, var1):
        if x < var1:
            r = 0
        else:
            r = 1
        return r

    @staticmethod
    def _loadBenchTable(bench_table):
        """
        Load benchmarking table and return DataFrame

        :param bench_table:
        :return:
        """
        fields = [benchmark.fd.GOID, benchmark.fd.FFRID, benchmark.fd.BENCH_SRC, benchmark.fd.Name_Expert]
        input = benchmark._load_streams(bench_table, fields)
        df = pd.DataFrame(input)

        return df

    @staticmethod
    def export_benchmarking_dom_results(bench_dom, stamp, writer, messages):
        """
        Now export benchmarking dominance stats

        :param bench_dom:
        :param stamp:
        :param writer:
        :return:
        """

        dom_bench_list = []
        # Simply resorting the columns of the data frame
        try:
            bench_dom_sorted = bench_dom[[
                "SCE_NAME", "FFRID", "Name_Expert", "Pressure", "NUM",
                "BENCH_SRC"]]
            for row in bench_dom_sorted._values:
                dom_bench_list.append([stamp] + [val for val in row])
        except Exception as err:
            arcpy.AddError ("Something went wrong with calculating"
                " the bench dom statistics. " + str(err))

        tools.export_excel(
            df=pd.DataFrame(dom_bench_list),
            name="Bench_dom",
            writer=writer)

# define fucntions from the original .../stats/backbone.py module
class backbone(object):

    def __init__(self):
        pass

    @staticmethod
    def backbone_stats(stream_array, scenario_name, min_length,
                    sta_pickle_folder, writer):
        """
        Calculate backbone stats

        :param stream_array:
        :param scenario_name:
        :param min_length:
        :param sta_pickle_folder:
        :param writer:
        :return:
        """

        # River_stats_1(two status options)
        bb0 = backbone.backbone_stats_0(scenario_name, stream_array)

        # River_stats_2 (three status options)
        bb1 = backbone.backbone_stats_1(scenario_name, stream_array)

        # River_stats_good
        bb2 = backbone.backbone_stats_2(scenario_name, stream_array)

        # List_of_FFRs
        bb3 = backbone.backbone_stats_3(scenario_name, stream_array, min_length)

        # Saving raw strings to output folder for later testing.
        for i, o in enumerate([bb1, bb0, bb3, bb2]):
            # https://stackoverflow.com/a/522578/344647
            tools.save_as_cpickle(pickle_object=o,
                                folder=sta_pickle_folder,
                                name=scenario_name + "bb" + str(i),
                                file_extension=".bb")

        tools.export_excel(bb0, 'River_stats_1', writer)
        tools.export_excel(bb1, 'River_stats_2', writer)
        tools.export_excel(bb2, 'River_stats_good', writer, True)

        # add the min_length to the end of the sheet name to indicate that only rivers above the min_length included. 
        tools.export_excel(bb3, 'List_of_FFRs_'+str(min_length), writer)

    @staticmethod
    def backbone_stats_0(sce_name, stream_array):
        """
        Calculating backbone statistics for Table 1.
        Only two categories: free-flowing or not (no "good" status)

        :param sce_name:
        :param stream_array:
        :return:
        """

        # Converting to panda data frame
        df = pd.DataFrame(stream_array)
        df = df.loc[df["INC"] == 1]

        df.loc[:, "SCE"] = sce_name

        # Calculating a field for "Length Category" and thus grouping the
        # backbone rivers into five categories:
        #
        # very short 0-10 km (not analyzed)
        # short rivers: 10 - 100 km
        # medium rivers 100 - 500 km
        # long rivers 500 - 1000 km
        # very long rivers > 1000 km

        df.loc[:, sce_name + "_LCAT"] = df[benchmark.fd.BB_LEN_KM].apply(backbone.get_length_cat)

        # Function to group by continent, river, length category,
        # and free-flowing status and calculate length (km), volume (million
        # cubic meters) and connectivity to ocean.
        # see Table 1

        fun1 = {'SCE': 'first',
                benchmark.fd.LENGTH_KM: np.sum,
                benchmark.fd.VOLUME_TCM: np.sum,
                benchmark.fd.BB_OCEAN: 'max'}

        tbl_1_temp = df.groupby([benchmark.fd.CON_ID,
                                benchmark.fd.BB_ID,
                                sce_name + "_LCAT",
                                sce_name + "_FF1"],
                                as_index=False).agg(fun1)

        # The result is a list of rivers by continent. We're adding a "count"
        # field (NUM) so we can now count the number of rivers
        tbl_1_temp.loc[:, "NUM"] = 1

        # Then repeat the same function as above

        fun2 = {'SCE': 'first',
                benchmark.fd.LENGTH_KM: np.sum,
                benchmark.fd.VOLUME_TCM: np.sum,
                benchmark.fd.BB_OCEAN: np.sum,
                "NUM": np.sum}

        tbl_1 = tbl_1_temp.groupby([benchmark.fd.CON_ID,
                                    sce_name + "_LCAT",
                                    sce_name + "_FF1"],
                                as_index=False).agg(fun2)

        # Some cleanup
        tbl_1.rename(columns={sce_name + "_LCAT": 'LCAT'}, inplace=True)
        tbl_1.rename(columns={sce_name + "_FF1": 'CAT_FFR'}, inplace=True)

        return tbl_1

    @staticmethod
    def backbone_stats_1(sce_name, stream_array):
        """
        Calculating backbone statistics for Table 1.
        Three categories presented in paper: free-flowing, good status, impacted

        :param sce_name:
        :param stream_array:
        :return:
        """

        # Converting to panda data frame
        df = pd.DataFrame(stream_array)
        df = df.loc[df["INC"] == 1]

        df.loc[:, "SCE"] = sce_name

        # Calculating a field for "Length Category" and thus grouping the
        # backbone rivers into five categories:
        #
        # very short 0-10 km (not analyzed)
        # short rivers: 10 - 100 km
        # medium rivers 100 - 500 km
        # long rivers 500 - 1000 km
        # very long rivers > 1000 km

        df.loc[:, sce_name + "_LCAT"] = df[benchmark.fd.BB_LEN_KM].apply(backbone.get_length_cat)

        # Function to group by continent, river, length category,
        # and free-flowing status and calculate length (km), volume (million
        # cubic meters) and connectivity to ocean.
        # see Table 1

        fun1 = {'SCE': 'first',
                benchmark.fd.LENGTH_KM: np.sum,
                benchmark.fd.VOLUME_TCM: np.sum,
                benchmark.fd.BB_OCEAN: 'max'}

        tbl_1_temp = df.groupby([benchmark.fd.CON_ID,
                                benchmark.fd.BB_ID,
                                sce_name + "_LCAT",
                                sce_name + "_FF2"],
                                as_index=False).agg(fun1)

        # The result is a list of rivers by continent. We're adding a "count"
        # field (NUM) so we can now count the number of rivers
        tbl_1_temp.loc[:, "NUM"] = 1

        # Then repeat the same function as above

        fun2 = {'SCE': 'first',
                benchmark.fd.LENGTH_KM: np.sum,
                benchmark.fd.VOLUME_TCM: np.sum,
                benchmark.fd.BB_OCEAN: np.sum,
                "NUM": np.sum}

        tbl_1 = tbl_1_temp.groupby([benchmark.fd.CON_ID,
                                    sce_name + "_LCAT",
                                    sce_name + "_FF2"],
                                as_index=False).agg(fun2)

        # Some cleanup
        tbl_1.rename(columns={sce_name + "_LCAT": 'LCAT'}, inplace=True)
        tbl_1.rename(columns={sce_name + "_FF2": 'CAT_FFR'}, inplace=True)

        return tbl_1

    @staticmethod
    def backbone_stats_2(sce_name, stream_array):
        """
        Calculating backbone statistics for Table 1.
        Three categories presented in paper: free-flowing, good status, impacted

        :param sce_name:
        :param stream_array:
        :return:
        """

        # Converting to panda data frame
        df = pd.DataFrame(stream_array)
        df = df[df["INC"] == 1]

        df.loc[:, "SCE"] = sce_name

        # Select only rivers in "good" status
        df = df[df[sce_name + "_FF2"] == 2]

        # Function to group by continent, river, length category,
        # and free-flowing status and calculate length (km), volume (million
        # cubic meters) and connectivity to ocean.
        # see Table 1

        fun1 = {'SCE': 'first',
                benchmark.fd.LENGTH_KM: np.sum,
                benchmark.fd.VOLUME_TCM: np.sum
                }

        good_temp = df.groupby([benchmark.fd.CON_ID,
                                benchmark.fd.BB_ID],
                            as_index=False).agg(fun1)

        good_temp.loc[:, sce_name + "_LCAT"] = \
            good_temp[benchmark.fd.LENGTH_KM].apply(backbone.get_length_cat)

        # The result is a list of rivers by continent. We're adding a "count"
        # field (NUM) so we can now count the number of rivers
        good_temp.loc[:, "NUM"] = 1

        # Then repeat the same function as above

        fun2 = {'SCE': 'first',
                benchmark.fd.LENGTH_KM: np.sum,
                benchmark.fd.VOLUME_TCM: np.sum,
                "NUM": np.sum}

        good = good_temp.groupby([benchmark.fd.CON_ID, sce_name + "_LCAT"],
                                as_index=False).agg(fun2)

        # Some cleanup
        good.rename(columns={sce_name + "_LCAT": 'LCAT'}, inplace=True)

        return good

    @staticmethod
    def backbone_stats_3(scenario_name, stream_array, min_length):
        """
        Calculating backbone statistics for Excel appendix, i.e. list
        of free-flowing rivers larger than 500 km.

        :param scenario_name:
        :param stream_array:
        :param min_length: minimum threshold length to be analyzed. For global
        analysis, it is 500km. More regional analysis could have a lower
        threshold, since there might not be rivers larger than 500 km

        :return:
        """

        # Converting to panda data frame
        df = pd.DataFrame(stream_array)
        df = df.loc[df["INC"] == 1]

        df.loc[:, "SCE"] = scenario_name

        # Aggregating length of each bb river...
        fct = {"SCE": 'first', benchmark.fd.CON_ID: 'first', benchmark.fd.BAS_NAME: 'first',
            benchmark.fd.LENGTH_KM: np.sum, benchmark.fd.VOLUME_TCM: np.sum,
            benchmark.fd.BB_NAME: 'first', benchmark.fd.BB_OCEAN: 'first', benchmark.fd.ISO_NAME: 'first',
            benchmark.fd.BB_LEN_KM: 'first', benchmark.fd.RIV_ORD: np.min,
            scenario_name + "_FF1": 'max',
            scenario_name + "_FF2": 'max',
            scenario_name + "_FFID": 'first',
            }
        length = df.groupby([benchmark.fd.BB_ID], as_index=False).agg(fct)
        length.loc[:, "NUM"] = 1

        # Result 1: Creating a list of FFR larger than X (min_length) km
        ff_river_list = length[length[benchmark.fd.LENGTH_KM] > min_length]

        # Only select free-flowing rivers
        ff_river_list = ff_river_list[ff_river_list[scenario_name + "_FF1"] == 1]

        # Select a subset of fields
        ff_river_list = ff_river_list[["SCE", benchmark.fd.CON_ID, benchmark.fd.BAS_NAME, benchmark.fd.BB_ID,
                                    benchmark.fd.BB_NAME, benchmark.fd.ISO_NAME, benchmark.fd.LENGTH_KM, benchmark.fd.RIV_ORD,
                                    benchmark.fd.BB_OCEAN, benchmark.fd.BAS_NAME]]

        return ff_river_list

    @staticmethod
    def get_length_cat(x):
        r = 0

        if x >= -1:
            r = 0
        if x > 10:
            r = 10
        if x > 100:
            r = 100
        if x > 500:
            r = 500
        if x > 1000:
            r = 1000

        return r


########## scripts functions ##########
# define functions from the original .../scripts/ffr_run_csi.py module
class ffr_run_csi(object):

    fd = var

    def __init__(self):
        pass

    @staticmethod
    def run_csi(stamp, para, scenarios, st_flds, paths, messages):
        """
        This is the main function to calculate the Connectivity Status Index (CSI); to calculate the river
        status, and to post-process the results into tables.

        The module is divided into three parts:

        1) Calculation of the CSI for each scenario.
        The fields that hold the CSI as well as the Dominance are added
        for each scenario and are named after the scenario name given
        in the config file

        2) The calculation of river status, in field ``CAT_FFR`` which
        determines the river or river stretch as either *'Free-flowing' (1)*,
        with *'Good connectivity status' (2)*, or *'impacted' (3)

        3) The calculation of benchmark, global and sensitivity statistics.

        :param stamp: Timestamp
        :param para: Parameters
        :param scenarios: Scenario set
        :param st_flds: list of fields
        :param paths: path settings
        :return:

        """

        # Looping through the individual scenarios
        for scenario in scenarios:

            sce_name = scenario[0]
            list_of_fields = scenario[1]
            list_of_weights = scenario[2]
            csi_threshold = scenario[3]
            flood_weight_damp = scenario[4]
            filter_thres = scenario[5]
            to_process = scenario[6]
            to_export = scenario[7]

            if to_process == 'NO':
                ffr_run_csi.prt("Skipped: " + sce_name, messages)
                continue
            else:
                ffr_run_csi.prt("Processing: " + sce_name, messages)

            # Define output CSI table
            csi_tb = paths["gdb_full_path"] + "\\" + "csi_tb"

            # Define output CSI fc
            csi_fc_name = "csi_fc_" + str(sce_name)

            stream_array = tools.load_stream_array(
                stream_feature_class=para["streams_fc"],
                stream_fields=st_flds,
                messages = messages)

            # Adding results fields to output table

            # Get the names of new csi fields to append
            sce_name, dom_field_name, ff_field_name, \
            csi_field_names = tools.get_csi_field_names(name=sce_name)

            # Get the names of new ffr fields to append
            ffr_stat1_field, ffr_stat2_field, ffr_dis_field, \
            ffr_field_names = tools.get_ffr_field_names(name=sce_name)

            ffr_run_csi.prt("Adding results fields to stream array", messages)
            stream_csi = tools.add_fields(array=stream_array,
                                        desc=csi_field_names +
                                            ffr_field_names)

            ffr_run_csi.prt("", messages)
            ffr_run_csi.prt("***********************", messages)
            ffr_run_csi.prt("PART 1: Calculating CSI", messages)
            ffr_run_csi.prt("***********************", messages)
            ffr_run_csi.prt("", messages)

            

            stream_csi = csi.calculate_csi(
                streams_array=stream_csi,
                csi_field_name=sce_name,
                dom_field_name=dom_field_name,
                ff_field_name=ff_field_name,
                fields=list_of_fields,
                weights=list_of_weights,
                flood_weight=flood_weight_damp,
                csi_threshold=csi_threshold,
                test_pickle_folder=paths["test_pickle_folder"], 
                messages = messages)

            # Saving CSI slice to Pickle for later conducting sensitivity analysis
            # Each scenario result will have their own pickle. Sensitivity
            # analysis loads the pickles and processes them together
            tools.save_as_cpickle(pickle_object=stream_csi[sce_name],
                                folder=paths["sta_csi_folder"], name=sce_name,
                                file_extension=".csi")

            # Assemble a results list that holds attributes of the scenario run_sed
            # Results from the global analysis will later be added. After each
            # scenario is run_sed, the list gets added to a list of lists. The list
            # of lists becomes the sheet "Global_stats" in the results excel
            result_list = [stamp]
            result_list += [sce_name]
            for f in list_of_fields:
                result_list += [f]
            for w in list_of_weights:
                result_list += [w]
            result_list += [csi_threshold] + \
                        [flood_weight_damp, to_process, filter_thres]

            ffr_run_csi.prt("", messages)
            ffr_run_csi.prt("*************************************", messages)
            ffr_run_csi.prt("PART 2: Calculating global statistics", messages)
            ffr_run_csi.prt("*************************************", messages)
            ffr_run_csi.prt("", messages)

            global_stats = gstat.post_stats_global_single(
                stream_csi, sce_name, csi_threshold)

            # The results of the global global_stats.py analysis is appended to the
            # results list
            for item in global_stats:
                result_list.append(item)

            ffr_run_csi.prt("", messages)
            ffr_run_csi.prt("***********************************************", messages)
            ffr_run_csi.prt("PART 3: Calculating global dominance statistics", messages)
            ffr_run_csi.prt("***********************************************", messages)
            ffr_run_csi.prt("", messages)
            dom_stats = dom.post_stats_dom_single(
                stream_csi, sce_name, csi_threshold)

            dom_stats["Stamp"] = stamp
            dom_stats_sort = dom_stats[["Stamp", "SCE_NAME", "Pressure", "NUM"]]

            tools.export_excel(dom_stats_sort, "Global_dom", paths["writer"],
                            False)

            ffr_run_csi.prt("", messages)
            ffr_run_csi.prt("**********************************", messages)
            ffr_run_csi.prt("PART 4: Dissolving Backbone Rivers", messages)
            ffr_run_csi.prt("**********************************", messages)
            ffr_run_csi.prt("", messages)

            # Filtering, dissolving and aggregating
            # This part is creates a copy of the stream array, applies filtering
            # and dissolving operations and then overwrites the FFR status fields,
            # and well as the river stretch IDs. The CSI value remains as is.

            # Make a copy of original results.
            stream_alt = np.copy(stream_csi)

            # Dissolve
            ffr_run_csi.prt("Dissolving part 1 of %s: " % ff_field_name, messages)
            stream_alt = sta.dissolve_rivers(stream_alt, ff_field_name,
                                            ffr_dis_field)

            # The spatial dissolving identifies river stretches that were both small and had a
            # disproportionally high impact on the CSI. This function conducts a spatial selected
            # of the river reaches that caused the sections with the high impact
            ffr_run_csi.prt("Apply filter for %s: " % ff_field_name, messages)
            bb_ids_to_filter = sta.apply_volume_filter(
                csi_fc=stream_alt,
                ff_field=ff_field_name,
                dis_id_field=ffr_dis_field,
                pct_aff_thres=filter_thres, 
                messages = messages)

            # Spatial selection and overwrite with zeros
            stream_alt = sta.update_csi(stream_alt, bb_ids_to_filter, ffr_dis_field, sce_name, ff_field_name, messages)

            # Dissolve again
            ffr_run_csi.prt("Dissolving part 2 of %s: " % ff_field_name, messages)
            stream_alt = sta.dissolve_rivers(stream_alt, ff_field_name,
                                            ffr_dis_field)

            ffr_run_csi.prt("Updating array %s: " % ff_field_name, messages)
            stream_csi = sta.update_streams_with_diss_id(
                stream_csi, stream_alt, ffr_dis_field)

            # Status calculations
            ffr_run_csi.prt("Calculating Status of %s: " % ff_field_name, messages)
            stream_csi = sta.calculate_sta(stream_csi, stream_alt, ff_field_name, ffr_stat1_field,
                                        ffr_stat2_field, ffr_dis_field, messages)

            ffr_run_csi.prt("", messages)
            ffr_run_csi.prt("*******************************************", messages)
            ffr_run_csi.prt("PART 5: Calculating benchmarking statistics", messages)
            ffr_run_csi.prt("*******************************************", messages)
            ffr_run_csi.prt("", messages)

            bench_val, bench_dom = benchmark.post_stats_bench_single(
                stream_array_mod=stream_alt,
                scenario_name=sce_name,
                bench_fc=para["bench_fc"],
                csi_threshold=csi_threshold,
                messages = messages)

            # Adding the value for number of free-flowing rivers at the end of
            # the global results list
            result_list += [bench_val]

            gstat.export_global_stats_results_to_excel(name_sheet="Global_stats",
                                                    result_list=result_list,
                                                    writer=paths["writer"])

            benchmark.export_benchmarking_dom_results(bench_dom=bench_dom,
                                                stamp=stamp,
                                                writer=paths["writer"],
                                                messages = messages)

            ffr_run_csi.prt("", messages)
            ffr_run_csi.prt("***************************************", messages)
            ffr_run_csi.prt("PART 6: Calculating backbone statistics", messages)
            ffr_run_csi.prt("***************************************", messages)
            ffr_run_csi.prt("", messages)

            backbone.backbone_stats(stream_csi, sce_name, para["min_length"],
                            paths["sta_pickle_folder"], paths["writer"])

            ffr_run_csi.prt("", messages)
            ffr_run_csi.prt("***************************************", messages)
            ffr_run_csi.prt("PART 7: Exporting results              ", messages)
            ffr_run_csi.prt("***************************************", messages)
            ffr_run_csi.prt("", messages)

            if to_export == 'YES':
                csi_table = str(csi_tb) + "_" +  str(sce_name)

                # Reduce numpy array to only necessary fields, i.e
                # get the names of new csi fields to append to input streams
                # feature class
                distilled_fields = [ffr_run_csi.fd.GOID,
                                    sce_name,
                                    dom_field_name,
                                    ff_field_name,
                                    ffr_stat1_field,
                                    ffr_stat2_field,
                                    ffr_dis_field]

                # rename #
                ffr_run_csi.prt("Renaming fields", messages)
                
                distilled_np = stream_csi[distilled_fields]
                
                df_distilled = pd.DataFrame(distilled_np)
                df_distilled.rename(columns={sce_name: 'CSI',
                                            dom_field_name: 'CSI_D', 
                                            ff_field_name: 'CSI_FF', 
                                            ffr_stat1_field: 'CSI_FF1', 
                                            ffr_stat2_field: 'CSI_FF2', 
                                            ffr_dis_field: 'CSI_FFID'}, inplace=True)  
                
                distilled = tools.pd_to_np(df_distilled)

                ffr_run_csi.prt("Exporting table: " + str(csi_table), messages)
                arcpy.da.NumPyArrayToTable(distilled, csi_table)

                ffr_run_csi.prt("Joining and exporting feature class", messages)

                output_fc = tools.export_joined(
                    output_geodatabase_path=paths["gdb_full_path"],
                    output_table_name=csi_fc_name,
                    table_to_join=csi_table,
                    join_table=para["streams_fc"],
                    messages = messages)

                ffr_run_csi.prt("Renaming fields", messages)
                tools.remove_csi_traces(output_fc, sce_name)
                ffr_run_csi.prt("Deleting join fields", messages)
                tools.delete_field(output_fc, ["OBJECTID_1", "GOID_1"], messages)




                ###### Create new MXD file to display the results #####
                
                # Get path to toolbox
                path = os.path.realpath(__file__)
                
                # split toolbox path into basepath and filename
                basepath, filename = os.path.split(path)

                # create path to template.mxd stored in same folder as toolbox 
                template_mxd = os.path.join(basepath, 'template.mxd')

                #  get data source path and dam feature class file name
                dams_source, dams_name = os.path.split(para["dams_fc"])
                
                # get data source path and dam feature class file name
                bench_source, bench_name = os.path.split(para["bench_fc"])

                # check if there is a .mxd file in the source directory matching the name of the template_mxd
                if arcpy.Exists(template_mxd):
                    tools.prt(messages,"Creating output ArcMap document for {}.".format(csi_fc_name))
                    
                    ### Customizing The Map Template ###
                    # create a list of lists with each sub-list containing the layer longName, data source, file name, and definition query
                    # we need to use the layer longName because a layer's longName will include the group layer name as part of the name.
                    # for the layer names provided in the template_mxd
                    template_data_list = [
                                          ["Barriers\Included Barriers", dams_source, dams_name, str(para["barrier_inc_field"]) + ' = 1' ],
                                          ["Barriers\Excluded Barriers", dams_source, dams_name, str(para["barrier_inc_field"]) + ' = 0' ],
                                          [r"MAP DOM - IMP\1",  paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 1 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\2", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 2 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\3", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 3 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\4", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 4 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\5", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 5 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\6", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 6 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\7", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 7 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\8", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 8 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\9", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 9 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\10", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 10 AND CSI < " + str(csi_threshold) ],
                                          [r"MAP DOM - IMP\Free-flowing", paths["gdb_full_path"], csi_fc_name, None],
                                          [r"MAP DOM - ALL\1", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 1" ],
                                          [r"MAP DOM - ALL\2", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 2" ],
                                          [r"MAP DOM - ALL\3", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 3" ],
                                          [r"MAP DOM - ALL\4", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 4" ],
                                          [r"MAP DOM - ALL\5", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 5" ],
                                          [r"MAP DOM - ALL\6", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 6" ],
                                          [r"MAP DOM - ALL\7", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 7" ],
                                          [r"MAP DOM - ALL\8", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 8" ],
                                          [r"MAP DOM - ALL\9", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 9" ],
                                          [r"MAP DOM - ALL\10", paths["gdb_full_path"], csi_fc_name, "RIV_ORD = 10" ],
                                          [r"MAP DOM - ALL\Free-flowing", paths["gdb_full_path"], csi_fc_name, None],
                                          [r"MAP CSI\1", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 1" ],
                                          [r"MAP CSI\2", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 2" ],
                                          [r"MAP CSI\3", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 3" ],
                                          [r"MAP CSI\4", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 4" ],
                                          [r"MAP CSI\5", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 5" ],
                                          [r"MAP CSI\6", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 6" ],
                                          [r"MAP CSI\7", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 7" ],
                                          [r"MAP CSI\8", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 8" ],
                                          [r"MAP CSI\9", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 9" ],
                                          [r"MAP CSI\10", paths["gdb_full_path"], csi_fc_name, "CSI < 100 AND RIV_ORD = 10" ],
                                          [r"MAP CSI\Free-flowing", paths["gdb_full_path"], csi_fc_name, "CSI = 100" ],
                                          [r"MAP FFR STATUS\IMP\very long", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 3 AND BB_LEN_KM > 1000" ],
                                          [r"MAP FFR STATUS\IMP\long", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 3 AND (BB_LEN_KM > 500 AND BB_LEN_KM < 1000)" ],
                                          [r"MAP FFR STATUS\IMP\medium", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 3 AND (BB_LEN_KM > 100 AND BB_LEN_KM < 500)" ],
                                          [r"MAP FFR STATUS\IMP\short", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 3 AND BB_LEN_KM < 100" ],
                                          [r"MAP FFR STATUS\GOOD\very long", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 2 AND BB_LEN_KM > 1000" ],
                                          [r"MAP FFR STATUS\GOOD\long", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 2 AND (BB_LEN_KM > 500 AND BB_LEN_KM < 1000)" ],
                                          [r"MAP FFR STATUS\GOOD\medium", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 2 AND (BB_LEN_KM > 100 AND BB_LEN_KM < 500)" ],
                                          [r"MAP FFR STATUS\GOOD\short", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 2 AND BB_LEN_KM < 100" ],
                                          [r"MAP FFR STATUS\FREE\very long", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 1 AND BB_LEN_KM > 1000" ],
                                          [r"MAP FFR STATUS\FREE\long", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 1 AND (BB_LEN_KM > 500 AND BB_LEN_KM < 1000)" ],
                                          [r"MAP FFR STATUS\FREE\medium", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 1 AND (BB_LEN_KM > 100 AND BB_LEN_KM < 500)" ],
                                          [r"MAP FFR STATUS\FREE\short", paths["gdb_full_path"], csi_fc_name, "CSI_FF2 = 1 AND BB_LEN_KM < 100" ],
                                          [r"Benchmark Rivers", bench_source, bench_name, None ],
                                          [r"MAP DOF\1", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 1" ],
                                          [r"MAP DOF\2", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 2" ],
                                          [r"MAP DOF\3", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 3" ],
                                          [r"MAP DOF\4", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 4" ],
                                          [r"MAP DOF\5", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 5" ],
                                          [r"MAP DOF\6", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 6" ],
                                          [r"MAP DOF\7", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 7" ],
                                          [r"MAP DOF\8", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 8" ],
                                          [r"MAP DOF\9", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 9" ],
                                          [r"MAP DOF\10", paths["gdb_full_path"], csi_fc_name, "DOF > 0 AND RIV_ORD = 10" ],
                                          [r"MAP DOF\Free-flowing", paths["gdb_full_path"], csi_fc_name, "DOF = 0" ],
                                          [r"MAP DOR\1", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 1" ],
                                          [r"MAP DOR\2", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 2" ],
                                          [r"MAP DOR\3", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 3" ],
                                          [r"MAP DOR\4", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 4" ],
                                          [r"MAP DOR\5", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 5" ],
                                          [r"MAP DOR\6", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 6" ],
                                          [r"MAP DOR\7", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 7" ],
                                          [r"MAP DOR\8", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 8" ],
                                          [r"MAP DOR\9", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 9" ],
                                          [r"MAP DOR\10", paths["gdb_full_path"], csi_fc_name, "DOR >0 AND RIV_ORD = 10" ],
                                          [r"MAP DOR\Free-flowing", paths["gdb_full_path"], csi_fc_name, "DOR = 0" ],
                                          [r"MAP SED\1", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 1" ],
                                          [r"MAP SED\2", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 2" ],
                                          [r"MAP SED\3", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 3" ],
                                          [r"MAP SED\4", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 4" ],
                                          [r"MAP SED\5", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 5" ],
                                          [r"MAP SED\6", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 6" ],
                                          [r"MAP SED\7", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 7" ],
                                          [r"MAP SED\8", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 8" ],
                                          [r"MAP SED\9", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 9" ],
                                          [r"MAP SED\10", paths["gdb_full_path"], csi_fc_name, "SED > 0 AND RIV_ORD = 10" ],
                                          [r"MAP SED\Free-flowing", paths["gdb_full_path"], csi_fc_name, "SED = 0" ],
                                          [r"MAP USE\1", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 1" ],
                                          [r"MAP USE\2", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 2" ],
                                          [r"MAP USE\3", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 3" ],
                                          [r"MAP USE\4", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 4" ],
                                          [r"MAP USE\5", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 5" ],
                                          [r"MAP USE\6", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 6" ],
                                          [r"MAP USE\7", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 7" ],
                                          [r"MAP USE\8", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 8" ],
                                          [r"MAP USE\9", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 9" ],
                                          [r"MAP USE\10", paths["gdb_full_path"], csi_fc_name, "USE > 0 AND RIV_ORD = 10" ],
                                          [r"MAP USE\Free-flowing", paths["gdb_full_path"], csi_fc_name, "USE = 0" ],
                                          [r"MAP RDD\1", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 1" ],
                                          [r"MAP RDD\2", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 2" ],
                                          [r"MAP RDD\3", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 3" ],
                                          [r"MAP RDD\4", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 4" ],
                                          [r"MAP RDD\5", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 5" ],
                                          [r"MAP RDD\6", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 6" ],
                                          [r"MAP RDD\7", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 7" ],
                                          [r"MAP RDD\8", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 8" ],
                                          [r"MAP RDD\9", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 9" ],
                                          [r"MAP RDD\10", paths["gdb_full_path"], csi_fc_name, "RDD >0 AND RIV_ORD = 10" ],
                                          [r"MAP RDD\Free-flowing", paths["gdb_full_path"], csi_fc_name, "RDD = 0" ],
                                          [r"MAP URB\1", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 1" ],
                                          [r"MAP URB\2", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 2" ],
                                          [r"MAP URB\3", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 3" ],
                                          [r"MAP URB\4", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 4" ],
                                          [r"MAP URB\5", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 5" ],
                                          [r"MAP URB\6", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 6" ],
                                          [r"MAP URB\7", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 7" ],
                                          [r"MAP URB\8", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 8" ],
                                          [r"MAP URB\9", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 9" ],
                                          [r"MAP URB\10", paths["gdb_full_path"], csi_fc_name, "URB > 0 AND RIV_ORD = 10" ],
                                          [r"MAP URB\Free-flowing", paths["gdb_full_path"], csi_fc_name, "URB = 0" ],
                                          ]
                    
                    # folder to save the results mxd to
                    output_folder = var.output_folder
                    # name of the output results mxd, the stamp will be appended to this name 
                    new_mxd_name = sce_name
                    # run the create_mxd tool using the above parameters
                    new_path_to_mxd = tools.create_mxd(template_mxd,
                                                        new_mxd_name,
                                                        template_data_list,
                                                        stamp,
                                                        output_folder,
                                                        messages)
                    
                    ### zoom to layers in new mxd ###
                    # create arcpy mapping document object
                    results_mxd = arcpy.mapping.MapDocument(new_path_to_mxd)
                    # create object for the data frame in results mxd
                    results_df = arcpy.mapping.ListDataFrames(results_mxd)[0]
                    # get description of input streams fc
                    stream_desc = arcpy.Describe(para["streams_fc"])
                    # set the extent of the results dataframe to match the input stream feature class
                    results_df.extent = stream_desc.extent
                    # save the results mxd
                    results_mxd.save()

                    # open the results automatically
                    os.system('"%s"' % new_path_to_mxd)

                else:
                    # if there is not a .mxd file in the source directory matching the name of the template_mxd. print the warning
                    arcpy.AddWarning("There is no 'template.mxd' document in the tool's directory. Unable to create output ArcMap document of results.")

            stream_array = None
            stream_csi = None

        ffr_run_csi.prt("", messages)
        ffr_run_csi.prt("***************************************", messages)
        ffr_run_csi.prt("PART 8: Post processing sensitivity    ", messages)
        ffr_run_csi.prt("***************************************", messages)
        ffr_run_csi.prt("", messages)

        # conduct sensitivity analysis
        sens.pst_csi_calculations(paths["sta_csi_folder"], messages)

        # delete the test pickle folder from the output folder
        test_pickle_folder = os.path.join(output_folder, paths["test_pickle_folder"] )
        tools.delete_path(test_pickle_folder, messages)

        # delete the Stats pickle folder from the output folder
        stats_pickle_folder = os.path.join(output_folder, paths["sta_pickle_folder"] )
        tools.delete_path(stats_pickle_folder, messages)

        # delete the  

        ffr_run_csi.prt("", messages)
        ffr_run_csi.prt("***************************************", messages)
        ffr_run_csi.prt("PART 9: Open results Excel and end     ", messages)
        ffr_run_csi.prt("***************************************", messages)
        ffr_run_csi.prt("", messages)

        os.system('"%s"' % paths["excel_file"])
        
        tools.prt(messages,datetime.datetime.now())
        ffr_run_csi.prt("Done", messages)

    @staticmethod
    def prt(txt, messages):
        logging.info(txt)
        messages.AddMessage(txt)

# define functions from the original .../scripts/ffr_run_dof.py module
class ffr_run_dof(object):

    fd = var 

    def __init__(self):
        pass

    @staticmethod
    def run_dof(stamp, para, paths, messages):
        """
        Set up for multiprocessing. Creates a isolated processing environment,
        where each hydrological river basin is processed separately

        :param stamp: timestamp
        :param para: parameters
        :param paths: pathnames
        :return:
        """

        dam_fc = para["dams_fc"]
        streams_fc = para["streams_fc"]

        update_stream_mode = para["update_mode"]

        drf_upstream = para["drf_upstream"]
        drf_downstream = para["drf_downstream"]

        barrier_inc_field = para["barrier_inc_field"]

        dof_field = para["dof_field"]
        dof_mode = para["dof_mode"]

        use_dam_level_df = para["use_dam_level_df"]

        # overwrite var dfd and dfu values to be the user specified fields
        var.DFU = para["DFU_field"]
        var.DFD = para["DFD_field"]

        gdb_full_path = paths["gdb_full_path"]

        output_folder = var.output_folder
        

        scratch_ws = output_folder + r"\Scratch"

        tools.delete_path(scratch_ws, messages)
        tools.create_path(scratch_ws, messages)



        in_basins = list(ffr_run_dof.get_unique(dam_fc, barrier_inc_field, messages))

        tools.prt(messages,"Loading {}".format(str(streams_fc)))
        streams = ffr_run_dof.load_streams(streams_fc, dof_field, messages)
        dams_temp = ffr_run_dof.load_dams(dam_fc, barrier_inc_field, use_dam_level_df, messages)

        pooled = False

        if pooled:
            pool = multiprocessing.Pool(8)

            jobs = []
            i = 1
        
            #tools.prt(messages,"Starting analysis pooled")

            if use_dam_level_df == 'YES':
                tools.prt(messages,"Calculating DOF using dam level discharge range factors from fields: " + "'" + str(dof.fd.DFU) + "'" + "(upstream) and "+"'"+ str(dof.fd.DFD) + "'"+ "(downstream)")
            else:
                tools.prt(messages,"Calculating DOF using upstream discharge range factor: {} and downstream discharge range factor: {}".format(drf_upstream, drf_downstream) )

            for basin in in_basins:
                # https://stackoverflow.com/a/8533626/344647
                # Much faster than querying the global dataset on disk is
                # to load the global dataset into memory first and then query it
                # using numpy methods
                streams_sel = np.copy(streams[streams[ffr_run_dof.fd.BAS_ID] == basin])

                # Also faster than querying the feature dataset on disk,
                # it is better to to load it once and then
                # use numpy indexing to get the dams we want
                dams_sel = np.copy(dams_temp[dams_temp[ffr_run_dof.fd.BAS_ID] == basin])

                jobs.append(pool.apply_async(ffr_run_dof.run_basin, (streams_sel, 
                                                        dams_sel,
                                                        basin,
                                                        stamp + str(i),
                                                        scratch_ws,
                                                        dof_field,
                                                        drf_upstream,
                                                        drf_downstream,
                                                        dof_mode,
                                                        use_dam_level_df)))

                i += 1

            pool.close()
            pool.join()

            out_basin = [job.get() for job in jobs]

        else:

            jobs = []
            i = 1

            #tools.prt(messages,"Starting analysis unpooled")

            if use_dam_level_df == 'YES':
                tools.prt(messages,"Calculating DOF using dam level discharge range factors from fields: " + "'" + str(dof.fd.DFU) + "'" + "(upstream) and "+"'"+ str(dof.fd.DFD) + "'"+ "(downstream)")
            else:
                tools.prt(messages,"Calculating DOF using upstream discharge range factor: {} and downstream discharge range factor: {}".format(drf_upstream, drf_downstream) )



            for basin in in_basins:
                # https://stackoverflow.com/a/8533626/344647
                # Much faster than querying the global dataset on disk is
                # to load the global dataset into memory first and then query it
                # using numpy methods
                streams_sel = np.copy(streams[streams[ffr_run_dof.fd.BAS_ID] == basin])

                # Also faster than querying the feature dataset on disk,
                # it is better to to load it once and then
                # use numpy indexing to get the dams we want
                dams_sel = np.copy(dams_temp[dams_temp[ffr_run_dof.fd.BAS_ID] == basin])

                jobs.append(ffr_run_dof.run_basin(streams_sel,
                                    dams_sel,
                                    basin,
                                    stamp + str(i),
                                    scratch_ws,
                                    dof_field,
                                    drf_upstream,
                                    drf_downstream,
                                    dof_mode,
                                    use_dam_level_df, 
                                    messages))
                i += 1

            out_basin = [job for job in jobs]


        # Merge the temporary outputs
        tools.prt(messages,"Merging temporary outputs into output table %s ..." % gdb_full_path)

        i = 0
        tbl = {}
        for bas in out_basin:
            i += 1

            with open(bas, 'rb') as fp:
                tbl[i] = cPickle.load(fp)

        #merged = np.concatenate(tbl.values(), 1)
        merged = np.concatenate(tbl.values())


        df = pd.DataFrame(merged)

        # Turn panda to numpy
        # https://my.usgs.gov/confluence/display/cdi/pandas.DataFrame+to+ArcGIS+Table
        x = np.array(np.rec.fromrecords(df.values))
        names = df.dtypes.index.tolist()
        x.dtype.names = tuple(names)


        output_table_location = os.path.join(gdb_full_path, "dof")
    
        arcpy.da.NumPyArrayToTable(x, output_table_location)
        
        tabledescription = arcpy.Describe(output_table_location)

        tools.add_index(lyr=output_table_location, field_name="GOID", messages = messages)

        # Update automatically
        if update_stream_mode == 'YES':
            tools.prt(messages,"Updating dof values in database {} ".format(streams_fc))

            tools.copy_between(to_join_fc=streams_fc,
                            to_join_field="GOID",
                            IntoJoinField=dof_field,
                            FromJoinFC=output_table_location,
                            FromJoinField="GOID",
                            FromValueField=dof_field,
                            over_mode=True,
                            over_value=0, 
                            messages = messages)

        tools.delete_path(scratch_ws, messages)

    @staticmethod
    def run_basin(streams, dams, basin, stamp, scratchws, dof_field, drf_upstream,
                drf_downstream, mode, use_dam_level_df, messages):
        """
        Calculate DOF for all barriers in a specified river basin

        :param streams:
        :param dams:
        :param basin:
        :param stamp:
        :param scratchws:
        :param dof_field:
        :param drf_upstream:
        :param drf_downstream:
        :param mode:
        :param use_dam_level_df:
        :return:
        """

        # Setup isolated path and environment
        temp_out_folder = ffr_run_dof.set_environment(scratchws, basin, stamp, messages)

        # Update network ids for rivers and dams
        tools.update_stream_routing_index(streams)
        tools.update_dam_routing_index(dams, streams)

        # Calculate and write DOF into designated field
        dof.calculate_DOF(dams, streams, mode, dof_field,
                                drf_upstream, drf_downstream, use_dam_level_df, messages)

        # Export table to temporay geodatabase
        final_table = ffr_run_dof.export(streams, basin, temp_out_folder)

        # returns the path to table for later merging
        return final_table

    @staticmethod
    def get_unique(dam_table, inc_field, messages):
        """
        Calculates a list of unique river basins that need to be processed based
        on the barriers to be considered.

        :param dam_table: numpy array with dams to process
        :param inc_field: field to determine dams to include
        :return: List of river basins
        """

        flds = [ffr_run_dof.fd.BAS_ID, inc_field]
        where_clause = inc_field + ' = 1'
        where_clause = where_clause.replace("'", "")

        dams = arcpy.da.TableToNumPyArray(
            dam_table, flds, where_clause, null_value=-1)

        in_basins = np.unique(dams[ffr_run_dof.fd.BAS_ID])

        if 0 in in_basins:
            sys.exit(
                "One of the dams has a basin ID of 0. BAS_ID cannot be zero. "
                "Please provide a BAS_ID other than 0 for "
                "all dams ")

        return in_basins

    @staticmethod
    def set_environment(scratch_ws, basin, stamp, messages):
        """
        Creates isolated output paths for specific basin using timestamp in scratch workspace.

        :param scratch_ws: scratch workspace
        :param basin:  river basin to process
        :param stamp: timestamp
        :return: fully specified pathname
        """

        out_folder = "basin_" + str(basin) + "_" + str(stamp)
        fullpath = os.path.join(scratch_ws, out_folder)

        tools.create_path(fullpath, messages)
        return fullpath

    @staticmethod
    def load_dams(dam_table, inc_field, use_dam_level_df, messages):
        """
        This function loads from the database

        :param dam_table:
        :param inc_field:
        :param use_dam_level_df:
        :return: numpy array with dams
        """
        

        if use_dam_level_df == 'YES':
            DFU_field = var.DFU
            DFD_field = var.DFD
            flds = [ffr_run_dof.fd.BAS_ID, ffr_run_dof.fd.GOID, ffr_run_dof.fd.STOR_MCM, DFU_field, DFD_field,
                    inc_field]

        else:
            flds = [ffr_run_dof.fd.BAS_ID, ffr_run_dof.fd.GOID, ffr_run_dof.fd.STOR_MCM, inc_field]

        arcpy.ListFields(dam_table)
        tools.check_fields(dam_table, flds, messages)

        whereBClause2 = inc_field + ' > 0'
        whereBClause3 = whereBClause2.replace("'", "")

        dams = arcpy.da.TableToNumPyArray(
            dam_table, flds, whereBClause3, null_value=0)

        return dams

    @staticmethod
    def load_streams(stream_table, dof_field, messages):
        """
        Loads the streams and adds a field for holding the DOF values

        :param stream_table: numpy array representing the river reaches
        :param dof_field: field name to store DOF results
        :return:
        """
        flds = [ffr_run_dof.fd.BAS_ID, ffr_run_dof.fd.GOID, ffr_run_dof.fd.NOID, ffr_run_dof.fd.NDOID, ffr_run_dof.fd.NUOID, ffr_run_dof.fd.RIV_ORD,
                ffr_run_dof.fd.DIS_AV_CMS, ffr_run_dof.fd.HYFALL]

        tools.check_fields(stream_table, flds, messages)

        arr = arcpy.da.TableToNumPyArray(stream_table, flds, null_value=0)
        arr = tools.add_fields(arr, [(str(dof_field), 'f4')])
        arr[dof_field] = 0
        return arr

    @staticmethod
    def export(streams, basin, folder):
        suffix = ".bas"
        name = "out_" + str(basin)
        fullname = name + suffix

        fullpath = os.path.join(folder, fullname)

        tools.save_as_cpickle(pickle_object=streams,
                            folder=folder,
                            name=name,
                            file_extension=suffix)
        return fullpath

    @staticmethod
    def create_gdb_workspace(gdb_folder, gdb_name):
        """
        Creates a path and a geodatabase with timestamp as name

        :param stamp:
        :param gdb_folder:
        :param gdb_name:
        :return: full path and geodatabase name
        """

        if not os.path.exists(gdb_folder):
            os.makedirs(gdb_folder)

        gdb_file_name = gdb_name + ".gdb"
        gdb_full_path = gdb_folder + "\\" + gdb_file_name

        arcpy.CreateFileGDB_management(gdb_folder, gdb_file_name)

        return gdb_full_path, gdb_file_name

# define functions from the original .../scripts/ffr_run_dor.py module
class ffr_run_dor(object):

    fd = var

    def __init__(self):
        pass

    @staticmethod
    def run_dor(stamp, para, paths, messages):
        """
        Set up for multiprocessing. Creates a isolated processing environment,
        where each hydrological river basin is processed separately

        :param stamp: timestamp
        :param para: parameters
        :param paths: pathnames
        :return:
        """

        dams_fc = para["dams_fc"]
        streams_fc = para["streams_fc"]

        update_mode = para["update_mode"]

        barrier_inc_field = para["barrier_inc_field"]
        dor_field = para["dor_field"]

        gdb_full_path = paths["gdb_full_path"]

        output_folder = var.output_folder

        scratch_ws = output_folder + r"\Scratch"

        tools.create_path(scratch_ws, messages)

        in_basins = list(ffr_run_dor.get_unique(dams_fc, barrier_inc_field))

        tools.prt(messages,"Loading {}".format(str(streams_fc)))
        streams = ffr_run_dor.load_streams(streams_fc, dor_field, messages)
        dams_temp = ffr_run_dor.load_dams(dams_fc, barrier_inc_field, messages)

        pooled = False

        if pooled:
            pool = multiprocessing.Pool()

            jobs = []
            i = 1

            tools.prt(messages,"Starting analysis pooled")

            for basin in in_basins:
                # https://stackoverflow.com/a/8533626/344647
                # Much faster than querying the global dataset on disk is
                # to load the global dataset into memory first and then query it
                # using numpy methods
                streams_sel = np.copy(streams[streams[ffr_run_dor.fd.BAS_ID] == basin])

                # Also faster than querying the feature dataset on disk,
                # it is better to to load it once and then
                # use numpy indexing to get the dams we want
                dams_sel = np.copy(dams_temp[dams_temp[ffr_run_dor.fd.BAS_ID] == basin])

                jobs.append(pool.apply_async(ffr_run_dor.run_basin, (streams_sel,
                                                        dams_sel, 
                                                        basin,
                                                        stamp + str(i), 
                                                        scratch_ws, 
                                                        dor_field)))
                
                i += 1

            pool.close()
            pool.join()

            out_basin = []
            
            for job in jobs:
                tools.prt(messages,'job.get():')
                tools.prt(messages,job.get())
                out_basin.append(job.get())

        else:

            jobs = []
            i = 1

            tools.prt(messages,"Starting analysis unpooled")
            for basin in in_basins:
                # https://stackoverflow.com/a/8533626/344647
                # Much faster than querying the global dataset on disk is
                # to load the global dataset into memory first and then query it
                # using numpy methods
                streams_sel = np.copy(streams[streams[ffr_run_dor.fd.BAS_ID] == basin])

                # Also faster than querying the feature dataset on disk,
                # it is better to to load it once and then
                # use numpy indexing to get the dams we want
                dams_sel = np.copy(dams_temp[dams_temp[ffr_run_dor.fd.BAS_ID] == basin])

                jobs.append(
                    ffr_run_dor.run_basin(streams_sel, dams_sel, basin, stamp + str(i), scratch_ws, dor_field, messages))
                i += 1

            out_basin = [job for job in jobs]

        # Merge the temporary outputs
        tools.prt(messages,"Merging temporary outputs into output table %s ..." % gdb_full_path)

        i = 0
        tbl = {}
        for bas in out_basin:
            i += 1

            with open(bas, 'rb') as fp:
                tbl[i] = cPickle.load(fp)

        #merged = np.concatenate(tbl.values(), 1)
        merged = np.concatenate(tbl.values())

        df = pd.DataFrame(merged)

        # Turn panda to numpy
        # https://my.usgs.gov/confluence/display/cdi/pandas.DataFrame+to+ArcGIS+Table
        x = np.array(np.rec.fromrecords(df.values))
        names = df.dtypes.index.tolist()
        x.dtype.names = tuple(names)

        output_table_location = gdb_full_path + "\\" + "dor"

        arcpy.da.NumPyArrayToTable(x, output_table_location)
        
        tools.add_index(lyr=output_table_location, field_name="GOID", messages = messages)

        # Update automatically
        if update_mode == 'YES':
            tools.prt(messages,"Updating dor values in database {} ".format(streams_fc))

            tools.copy_between(to_join_fc=streams_fc,
                            to_join_field="GOID",
                            IntoJoinField=dor_field,
                            FromJoinFC=output_table_location,
                            FromJoinField="GOID",
                            FromValueField=dor_field,
                            over_mode=True,
                            over_value=0,
                            messages = messages)

        tools.delete_path(scratch_ws, messages)

    @staticmethod
    def run_basin(streams, dams, basin, stamp, scratchws, dor_field, messages):
        """
        Calculate DOR for all barriers in a specified river basin

        :param streams:
        :param dams:
        :param basin:
        :param stamp:
        :param scratchws:
        :param dof_field:
        :param drf_upstream:
        :param drf_downstream:
        :param mode:
        :param use_dam_level_df:
        :return:
        """
        # setup isolated path and environment
        scratch_gdb = ffr_run_dor.set_environment(scratchws, basin, stamp, messages)

        # update netwrok ids for rivers and dams
        tools.update_stream_routing_index(streams)
        tools.update_dam_routing_index(dams, streams)

        tools.prt(messages,"Calculating DOR for basin {}".format(str(basin)))

        # calculate and write DOR into designated field
        dor.calculate_dor(dams, streams, dor_field)

        # export table to temporary geodatabase
        final_table = ffr_run_dor.export(streams, basin, scratch_gdb)

        # Return only reference (path) to table
        return final_table

    @staticmethod
    def get_unique(dam_table, inc_field):
        """
        Calculates a list of unique river basins that need to be processed based
        on the barriers to be considered.

        :param dam_table: umpy array with dams to process
        :param inc_field: field to determine dams to include
        :return: List of river basins
        """

        flds = [ffr_run_dor.fd.BAS_ID, inc_field]
        whereBClause = inc_field + ' = 1'
        whereBClause = whereBClause.replace("'", "")

        dams = arcpy.da.TableToNumPyArray(
            dam_table, flds, whereBClause, null_value=-1)

        in_basins = np.unique(dams[ffr_run_dor.fd.BAS_ID])

        if 0 in in_basins:
            sys.exit(
                "One of the dams has a basin ID of 0. BAS_ID cannot be zero. "
                "Please provide a BAS_ID other than 0 for "
                "all dams ")

        return in_basins

    @staticmethod
    def set_environment(scratch_ws, basin, stamp, messages):
        """
        Creates isolated output paths for specific basin using timestamp in scratch workspace.

        :param scratch_ws: scratch workspace
        :param basin:  river basin to process
        :param stamp: timestamp
        :return: fully specified pathname
        """

        out_folder = "basin_" + str(basin) + "_" + str(stamp)
        fullpath = os.path.join(scratch_ws, out_folder)

        tools.create_path(fullpath, messages)
        return fullpath

    @staticmethod
    def load_dams(dam_table, inc_field, messages):
        """
        This function loads from the database

        :param dam_table:
        :param inc_field:
        :param use_dam_level_df:
        :return: numpy array with dams
        """
        
        flds = [ffr_run_dor.fd.BAS_ID, ffr_run_dor.fd.GOID, ffr_run_dor.fd.STOR_MCM, inc_field]



        tools.check_fields(dam_table, flds, messages)

        whereBClause2 = inc_field + ' > 0'
        whereBClause3 = whereBClause2.replace("'", "")

        dams = arcpy.da.TableToNumPyArray(dam_table, flds, whereBClause3, null_value=0)


        return dams

    @staticmethod
    def load_streams(stream_table, dor_field, messages):
        """
        Loads the streams and adds a field for holding the DOF values

        :param stream_table: numpy array representing the river reaches
        :param dor_field: field name to store DOR results
        :return:
        """

        flds = [ffr_run_dor.fd.BAS_ID, ffr_run_dor.fd.GOID, ffr_run_dor.fd.NOID, ffr_run_dor.fd.NDOID, ffr_run_dor.fd.NUOID, ffr_run_dor.fd.RIV_ORD,
                ffr_run_dor.fd.DIS_AV_CMS, ffr_run_dor.fd.HYFALL]

        tools.check_fields(stream_table, flds, messages)

        arr = arcpy.da.TableToNumPyArray(stream_table, flds, null_value=0)
        arr = tools.add_fields(arr, [(str(dor_field), 'f4')])
        arr[dor_field] = 0
        return arr

    @staticmethod
    def export(streams, basin, folder):
        suffix = ".bas"
        name = "out_" + str(basin)
        fullname = name + suffix

        fullpath = os.path.join(folder, fullname)

        tools.save_as_cpickle(pickle_object=streams,
                            folder=folder,
                            name=name,
                            file_extension=suffix)
        return fullpath

    @staticmethod
    def create_gdb_workspace(gdb_folder, gdb_name):
        """
        Creates a path and a geodatabase with timestamp

        :param stamp:
        :param gdb_folder:
        :param gdb_name:
        :return:
        """

        if not os.path.exists(gdb_folder):
            os.makedirs(gdb_folder)

        gdb_file_name = gdb_name + ".gdb"
        gdb_full_path = gdb_folder + "\\" + gdb_file_name

        arcpy.CreateFileGDB_management(gdb_folder, gdb_file_name)

        return gdb_full_path, gdb_file_name

# define functions from the original .../scripts/ffr_run_sed.py module
class ffr_run_sed(object):

    fd = var

    def __init__(self):
        pass

    @staticmethod
    def run_sed(para, paths, messages):
        """
        Script to calculate the Sediment Trapping Index (SED)

        :param para: input parameters and path names for executing the script
        :param paths: output pathnames
        :return:
        """
        streams_fc = para["streams_fc"]
        dams_fc = para["dams_fc"]
        svol_field = para["svol_field"]
        barrier_inc_field = para["barrier_inc_field"]
        lakes_fc = para["lakes_fc"]
        sed_field = para["sed_field"]
        out_gdb = paths["gdb_full_path"]

        streams = ffr_run_sed.load_streams(streams_fc)
        streams, convert_dict = ffr_run_sed.update_stream_routing_index(streams, messages)

        barriers = ffr_run_sed.load_barriers(dams_fc, convert_dict,
                                svol_field, barrier_inc_field, messages)
        dam_volu_dict = ffr_run_sed.barriers_calculate(barriers, svol_field)

        lakes = ffr_run_sed.load_lakes(lakes_fc, convert_dict, messages)

        small_lake_loss_dict, lake_volu_dict = sed.lakes_calculate(lakes, messages)

        streams = sed.calculate_sed(streams, dam_volu_dict, lake_volu_dict,
                                            small_lake_loss_dict, messages)

        ffr_run_sed.prt("Exporting results sediment table", messages)

        outtbl = ffr_run_sed.export_results_table(streams, out_gdb)

        # Adding indices helps with joining tables to geometry
        arcpy.AddIndex_management(outtbl, ffr_run_sed.fd.GOID, ffr_run_sed.fd.GOID, "UNIQUE", "ASCENDING")

        # Update original database
        if para["update_mode"] == 'YES':
            tools.prt(messages,"Updating SED values in database {} ".format(streams_fc))
            try:
                tools.copy_between( to_join_fc = streams_fc,
                                    to_join_field = ffr_run_sed.fd.GOID,
                                    IntoJoinField = sed_field,
                                    FromJoinFC = outtbl,
                                    FromJoinField = ffr_run_sed.fd.GOID,
                                    FromValueField = ffr_run_sed.fd.SED,
                                    over_mode = True,
                                    over_value = 0, 
                                    messages = messages)

            except Exception as e:
                arcpy.AddError (str(e))
                sys.exit(0)

    @staticmethod
    def load_streams(stream_table):
        """
        Loading stream network and adding fields

        :param stream_table:
        :return: stream array with necessary fields
        """

        # Existing fields to load
        flds = [ffr_run_sed.fd.GOID, ffr_run_sed.fd.NOID, ffr_run_sed.fd.NDOID, ffr_run_sed.fd.NUOID,
                ffr_run_sed.fd.DIS_AV_CMS, ffr_run_sed.fd.BAS_ID, ffr_run_sed.fd.UPLAND_SKM,
                ffr_run_sed.fd.ERO_YLD_TON]

        arr = arcpy.da.TableToNumPyArray(stream_table, flds)

        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_LSS_LKS_OT_NAT, 'f8')])
        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_LSS_LKS_IN_NAT, 'f8')])
        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_NAT_UP, 'f8')])
        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_NAT, 'f8')])

        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_LSS_LKS_OT_ANT, 'f8')])
        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_LSS_LKS_IN_ANT, 'f8')])
        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_LSS_DMS_ANT, 'f8')])
        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_ANT_UP, 'f8')])
        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_ANT, 'f8')])

        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED_LSS_TOT, 'f8')])

        arr = tools.add_fields(arr, [(ffr_run_sed.fd.SED, 'f8')])

        arr[ffr_run_sed.fd.SED_NAT] = 0
        arr[ffr_run_sed.fd.SED_NAT_UP] = 0
        arr[ffr_run_sed.fd.SED_LSS_LKS_OT_NAT] = 0
        arr[ffr_run_sed.fd.SED_LSS_LKS_IN_NAT] = 0

        arr[ffr_run_sed.fd.SED_ANT] = 0
        arr[ffr_run_sed.fd.SED_ANT_UP] = 0
        arr[ffr_run_sed.fd.SED_LSS_LKS_OT_ANT] = 0
        arr[ffr_run_sed.fd.SED_LSS_LKS_IN_ANT] = 0
        arr[ffr_run_sed.fd.SED_LSS_DMS_ANT] = 0

        arr[ffr_run_sed.fd.SED_LSS_TOT] = 0

        arr[ffr_run_sed.fd.SED] = 0

        return arr

    @staticmethod
    def update_stream_routing_index(streams, messages):
        """
        Function to sort the stream network using the upstream area. This allows
        the network to be processed in order from headwaters to the ocean. Afterwards
        the Network IDS are recalculated

        :param streams: numpy array of stream network
        :return:
        """

        tools.prt(messages,"Updating stream index")

        # Maintain the old GOID values in a new field
        streams = tools.add_fields(streams, [("OGOID", 'i4')])
        streams["OGOID"] = streams["GOID"]

        # Sort the array by upland area and basin
        # This is key to being able to process river network from top to bottom
        streams.sort(order=['BAS_ID', 'UPLAND_SKM'])

        # Create Routing Dictionaries and fill
        oid_dict = {}
        convert_dict = {}
        ups_dict = defaultdict(str)
        down_dict = defaultdict(long)

        i = 1
        for myrow in streams:
            oid_dict[int(myrow["GOID"])] = i
            i += 1

        i = 1
        for myrow in streams:
            dn_old_oid = myrow["NDOID"]
            new_oid = oid_dict.get(int(dn_old_oid), -1)

            if new_oid != -1:
                # Write OID of next downstream reach
                down_dict[i] = new_oid
                # Write OID of next upstream reach
                exi_value = ups_dict.get(int(new_oid), -99)

                if exi_value == -99:
                    ups_dict[int(new_oid)] = i
                else:
                    new_value = str(exi_value) + '_' + str(i)
                    ups_dict[int(new_oid)] = new_value
            i += 1

        # Writing index values back to numpy
        i = 1
        for myrow in streams:
            myrow["NOID"] = i
            myrow["NDOID"] = down_dict[i]
            myrow["NUOID"] = ups_dict[i]

            i = i + 1

        # Create Dictionary to convert old (KEY) to new (VALUE)
        i = 1
        for myrow in streams:
            old = myrow["OGOID"]
            new = myrow["NOID"]
            convert_dict[int(old)] = new
            i += 1

        return streams, convert_dict

    @staticmethod
    def load_lakes(lakes_table, convert_dict, messages):
        """
        Load lakes and add field for sediment trapping calculations

        :param lakes_table:
        :param convert_dict:
        :return: numpy array of lakes
        """

        tools.prt(messages,"Loading lakes")

        flds = ["GOID", "GOOD", "Lake_type", "SED_ACC", "IN_STREAM", "IN_CATCH",
                "Vol_total", "Dis_avg", "Res_time"]

        arr = arcpy.da.TableToNumPyArray(lakes_table, flds, null_value=0)

        arr = tools.add_fields(arr, [("TE_brune", 'f8')])
        arr = tools.add_fields(arr, [("LOSS_LKES_OUT_NET", 'f8')])

        arr["TE_brune"] = 0
        arr["LOSS_LKES_OUT_NET"] = 0

        for a in arr:
            a["GOID"] = convert_dict.get(a["GOID"], 0)

        return arr

    @staticmethod
    def load_barriers(barriers_table, convert_dict, svol_field, barrier_inc_field, messages):
        """
        Loading dams and converting old to new

        :param barriers_table:
        :param convert_dict:
        :param svol_field:
        :param barrier_inc_field:
        :return:
        """
        tools.prt(messages,"Loading dams")

        # Existing fields to load
        flds = [ffr_run_sed.fd.GOID, ffr_run_sed.fd.NOID, svol_field, barrier_inc_field]

        # Load the array from file
        arr = arcpy.da.TableToNumPyArray(barriers_table, flds, null_value=0)

        for a in arr:
            a[ffr_run_sed.fd.GOID] = convert_dict.get(a[ffr_run_sed.fd.GOID], 0)

        arr2 = arr[(arr[barrier_inc_field] == 1)]

        return arr2

    @staticmethod
    def barriers_calculate(barriers, svol_field):
        """
        In some cases, there are multiple reservoirs located on a river reach.
        This function calculates the sum of reservoir volume for each river reach.

        :param barriers:
        :param svol_field:
        :return:
        """
        dam_volu_dict = {}

        for b in barriers:
            goid = b[ffr_run_sed.fd.GOID]
            dam_volu_dict[goid] = dam_volu_dict.get(goid, 0) + b[svol_field]

        return dam_volu_dict

    @staticmethod
    def export_results_table(streams, out_gdb):
        out_tbl = out_gdb + "\\sed"
        arcpy.da.NumPyArrayToTable(streams[[ffr_run_sed.fd.GOID, ffr_run_sed.fd.SED]], out_tbl)
        return out_tbl

    @staticmethod
    def prt(txt, messages):
        logging.info(txt)
        messages.AddMessage(txt)


########## indices functions ##########
# define functions from the original .../indices/csi.py module
class csi(object):

    def __init__(self):
        pass

    @staticmethod
    def calculate_csi(streams_array, csi_field_name, dom_field_name,
                    ff_field_name, fields, weights, flood_weight,
                    csi_threshold, messages, test_pickle_folder=''):
        """
        Calculating the CSI and the DOM index

        :param streams_array: input stream feature class converted to numpy array
        :param csi_field_name: Field name that holds the CSI index
        :param dom_field_name: Field name that holds the DOM index
        :param ff_field_name: Field name that determines if CSI threshold is
        exceeded for the particular reach (0 = No or 1 = Yes)
        :param fields: List of fields to include in CSI calculations,
        as defined in Excel sheet. Fields must be present in the input stream file
        :param weights: list of weights (0-100%) corresponding to the
        list of fields
        :param flood_weight: Floodplain weight factor, see Excel sheet (=50%)
        :param csi_threshold: CSI threshold, defined in Excel sheet (=95%)
        :return:
        """

        # Extracting fields from numpy to individual panda series, makes for much
        # more efficient processing
        dof = pd.Series(streams_array[fields[0]])
        dor = pd.Series(streams_array[fields[1]])
        sed = pd.Series(streams_array[fields[2]])
        use = pd.Series(streams_array[fields[3]])
        rdd = pd.Series(streams_array[fields[4]])
        urb = pd.Series(streams_array[fields[5]])
        fld = pd.Series(streams_array["FLD"])

        fld = fld / 100.0



        """ Calculating CSI """

        # Adding floodplain weighting to road and urban only
        rdd = rdd + ((rdd * fld) * (flood_weight / 100))
        urb = urb + ((urb * fld) * (flood_weight / 100))

        # resetting to 100 in case weighting overshoots
        rdd[rdd > 100.0] = 100.0
        urb[urb > 100.0] = 100.0

        # Then regular weighting
        dof = dof * weights[0]
        dor = dor * weights[1]
        sed = sed * weights[2]
        use = use * weights[3]
        rdd = rdd * weights[4]
        urb = urb * weights[5]

        # very small values are set to zero
        dof[dof < 0.1] = 0
        dor[dor < 0.1] = 0
        sed[sed < 0.1] = 0
        use[use < 0.1] = 0
        rdd[rdd < 0.1] = 0
        urb[urb < 0.1] = 0

        # Weighted overlay operation
        csi = 100.0 - ((dof +
                        dor +
                        sed +
                        use +
                        rdd +
                        urb) / 100.0)

        # Rounding to 5 decimal places
        csi = np.around(csi, decimals=5)

        # Assigning CSI values to base array
        streams_array[csi_field_name] = csi

        # Determines if CSI is exceeded or not
        sel = np.where(streams_array[csi_field_name] < csi_threshold, 0, 1)
        streams_array[ff_field_name] = sel

        """ Calculating DOM """

        combine = pd.concat([dof, dor, sed, use, rdd, urb], axis=1)

        # Test if everything is zero (no clear dominant pressure), in which
        # case a '1' is assigned to the array, which leads to "NAN" in the later
        # step (at position 6)
        nan = np.where(np.sum(combine, axis=1) == 0, 1, 0)

        # Must be converted to DataFrame, same as others.
        nan = pd.Series(nan)

        # Combine again to get the index right
        combine2 = pd.concat([dof, dor, sed, use, rdd, urb, nan], axis=1)

        # Find index with highest value
        # Only the first index value will be returned. So if two indicators
        # have the same value, the first one in the list becomes the dominant
        # one. This slightly favours DOF instead of DOR and RDD instead of urb
        dom = combine2.idxmax(axis=1, skipna=True)

        dom[dom == 0] = "DOF"
        dom[dom == 1] = "DOR"
        dom[dom == 2] = "SED"
        dom[dom == 3] = "USE"
        dom[dom == 4] = "RDD"
        dom[dom == 5] = "URB"
        dom[dom == 6] = "NAN"

        streams_array[dom_field_name] = dom

        return streams_array

# define functions from the original .../indices/dof.py module
class dof(object):

    fd = var

    def __init__(self):
        pass

    @staticmethod
    def calculate_DOF(dams, streams, mode, dof_field, drf_upstream, drf_downstream,
                    use_dam_level_df, messages):
        """
        Calculates the degree of fragmentation (DOF) given a set of dams

        :param dams: numpy array including one or several dams
        :param streams: numpy array with a stream network
        :param mode: numeric value defining the selection of decay function
        :param dof_field: field to fill with DOF value
        :param drf_upstream: multiplier to indicate deviation from discharge
            occurring at barrier, usually 5(=0.5 orders of magnitude). Can be set
            individually for each dam, and for each direction
        :param drf_downstream: multiplier to indicate deviation from discharge
            occurring at barrier, usually 5(=0.5 orders of magnitude). Can be set
            individually for each dam, and for each direction
        :param use_dam_level_df: If true, looks for DFU and DFD fields,
            which hold discharge range factor values for upstream and
            downstream direction
        :return: stream array with updates DOF values
        """

        ndoid = streams["NDOID"].tolist()
        nuoid = streams["NUOID"].tolist()
        wfall = streams["HYFALL"].tolist()
        disch = streams["DIS_AV_CMS"].tolist()

        upstream_mode = mode
        downstream_mode = mode

        nodes = []
        downstream_reaches = set([])
        stop = set([])
                

        # First process everything upstream of dam
        for dam in dams:
            # added np.int64 due to error
            # TypeError: list indices must be integers, not numpy.float64            
            dam_goid = np.int64(dam[dof.fd.GOID])


            if use_dam_level_df == 'YES':
                drf_upstream = dam[dof.fd.DFU]
                drf_downstream = dam[dof.fd.DFD]
            else:
                drf_upstream = drf_upstream
                drf_downstream = drf_downstream

            if drf_upstream <= 1:
                # To prevent log10(1) = 0
                drf_upstream = 1.000000000000001
                if drf_upstream < 1:
                    arcpy.AddWarning ("discharge range factor (upstream) can not be "
                        "lower than 1. Setting drf_upstream to 1")
            if drf_downstream <= 1:
                # To prevent log10(1) = 0
                drf_downstream = 1.000000000000001
                if drf_downstream < 1:
                    arcpy.AddWarning ("discharge range factor (downstream) can not be "
                        "lower than 1. Setting drf_downstream to 1")

            nodes.append(dam_goid)  # Add the dam to the nodes list to process
            stop.clear()
            stop.add(dam_goid)
            downstream_reaches.clear()

            discharge_barrier_location = disch[dam_goid - 1]

            if discharge_barrier_location == 0:
                streams[dam_goid - 1][dof_field] = 100
                continue

            # discharge range factor, usually 10 (= one order of magnitude)
            dis_low = discharge_barrier_location / drf_upstream
            dis_hgh = discharge_barrier_location * drf_downstream
    

            # 1) First process everything upstream of dam
            while 1:
                new_nodes = []
                if len(nodes) == 0:
                    break
                for n in nodes:

                    # Add one or several constraints on how to rout
                    if n != -1 and n != '':

                        # Check if node has waterfall. If the waterfall is on the
                        # reach where the dam is, stop routing upstream
                        if wfall[n - 1] == 0:
                            # Check for conditions of discharge
                            if dis_hgh >= disch[n - 1] >= dis_low:
                                # declare segment as fragmented
                                discharge_local = disch[n - 1]

                                local_impact_score = dof.get_dof_up(
                                    discharge_local=discharge_local,
                                    discharge_barrier=discharge_barrier_location,
                                    upstream_mode=upstream_mode,
                                    dis_range_factor=drf_upstream
                                    , messages = messages)

                                if streams[n - 1][dof_field] <= local_impact_score or streams[n - 1][dof_field] == 0:
                                    streams[n - 1][dof_field] = local_impact_score


                                # if local_impact_score > 0:  # find upstream nodes
                                upstream_oid = nuoid[n - 1].split("_")
                                if len(upstream_oid) > 0:
                                    for up in upstream_oid:
                                        if up != '' and up > 0:
                                            new_nodes.append(int(up))


                nodes = new_nodes

            # 2) Then process downstream
            nodes.append(dam_goid)  # Add the dam to the nodes list to process

            while 1:
                new_nodes = []
                if len(nodes) == 0:
                    break
                for n in nodes:
                    if n != -1 and n != '':
                        # declare segment as fragmented if local discharge of
                        # reach is within upper and lower limits
                        discharge_local = disch[n - 1]
                        if dis_low <= discharge_local <= dis_hgh:
                            local_impact_score = dof.get_dof_down(
                                discharge_local=discharge_local,
                                discharge_barrier=discharge_barrier_location,
                                downstream_mode=downstream_mode,
                                dis_range_factor=drf_downstream,
                                messages = messages)
                            if streams[n - 1][dof_field] <= local_impact_score or \
                                    streams[n - 1][dof_field] == 0:
                                streams[n - 1][dof_field] = local_impact_score

                            # if local_impact_score > 0:  # find downstream nodes
                            downstream_oid = ndoid[n - 1]
                            if downstream_oid > 0:
                                new_nodes.append(int(downstream_oid))
                                downstream_reaches.add(downstream_oid)

                nodes = new_nodes


    @staticmethod
    def get_dof_down(discharge_local, discharge_barrier, downstream_mode,
                    dis_range_factor, messages):
        """
        Calculates DOF for downstream reaches

        :param discharge_local: discharge (CMS) at current river reach to be processed
        :param discharge_barrier: discharge (CMS) at river reach of barrier location
        :param downstream_mode: defines the function of gradual decay of effect in the downstream
        direction
        :param dis_range_factor: defines the ratio between discharge_barrier and
        discharge_local below which DOF effects are considered
        :return: degree of fragmentation (DOF) in percent
        """
        if discharge_local < discharge_barrier:
            # prevents DOF to become larger than 100
            # in special cases when e.g, discharge decreases downstream
            # and becomes larger than local discharge
            discharge_local = discharge_barrier  # reset discharge

        if downstream_mode == 1:  # log two order
            a = abs(math.log10(discharge_local) - math.log10(discharge_barrier))
            b = a * (100 / math.log10(dis_range_factor))
            x = 100 - b
        else:
            tools.prt(messages,"discharge mode undefined")
            sys.exit()

        # I some cases x can get out of bounds
        if x < 0:
            x = 0
        if x > 100:
            x = 100

        return x

    @staticmethod
    def get_dof_up(discharge_local, discharge_barrier, upstream_mode,
                dis_range_factor, messages):
        """
        Calculates DOF for upstream reaches

        :param discharge_local: discharge (CMS) at current river reach to be processed
        :param discharge_barrier: discharge (CMS) at river reach of barrier location
        :param upstream_mode: defines the function of gradual decay of effect in the upstream direction
        :param dis_range_factor: defines the ratio between discharge_barrier and
        discharge_local below which DOF effects are considered
        :return: degree of fragmentation (DOF) in percent
        """

        if discharge_local > discharge_barrier:
            discharge_local = discharge_barrier
            # prevents DOF to become larger than 100
            # in special cases when e.g, discharge increases upstream,
            # and becomes smaller than local discharge

        if upstream_mode == 1:  # log one order
            a = abs(math.log10(discharge_local) - math.log10(discharge_barrier))
            b = a * (100 / math.log10(dis_range_factor))
            x = 100 - b
        else:
            tools.prt(messages,"discharge mode undefined")
            sys.exit()

        # I some cases x can get out of bounds
        if x < 0:
            x = 0
        if x > 100:
            x = 100

        return x

# define functions from the original .../indices/dom.py module
class dom(object):

    def __init__(self):
        pass

    @staticmethod
    def post_stats_dom_single(stream_array, csi_name, threshold):
        panda_df = pd.DataFrame(stream_array)
        panda_df = panda_df[panda_df[benchmark.fd.INC] == 1]

        dom_field_name = csi_name + str("_D")
        dom_all = dom.calculate_dominance_all(
            panda_df, dom_field_name, csi_name, threshold)
        return dom_all

    @staticmethod
    def calculate_dominance_all(panda_df, dom_field_name, sce_name, thres):
        panda_df.loc[:, "SCE_NAME"] = sce_name
        panda_df.loc[:, "NUM"] = 1

        sel = panda_df[panda_df[sce_name] < thres]

        fun = {'NUM': np.sum}
        dom = sel.groupby([dom_field_name, 'SCE_NAME'], as_index=False).agg(fun)
        dom.rename(columns={dom_field_name: 'Pressure'}, inplace=True)
        return dom

# define functions from the original .../indices/dor.py module
class dor(object):

    fd = var

    def __init__(self):
        pass

    @staticmethod
    def calculate_dor(dams, streams, dor_field):  # List of OIDs
        """
        Updates the degree of regulation (DOR) index, given a
        set of dams.

        :param dams: numpy array of barriers and their attributes
        :param streams: numpy array of streams and their attributes
        :param dor_field: field to store the DOR values
        :return: stream array with updated DOR values
        """

        length = streams.shape[0]
        ndoid = streams["NDOID"].tolist()
        disch = streams["DIS_AV_CMS"].tolist()
        svol = [0] * length

        for dam in dams:

            routing_list = set([])
            # added np.int64 due to error
            # TypeError: list indices must be integers, not numpy.float64              
            dam_oid = np.int64(dam[dor.fd.GOID])

            while 1:
                new_node = -1
                if dam_oid == -1:
                    break

                if dam_oid not in routing_list:
                    routing_list.add(dam_oid)

                    # Calculate DOR and update array
                    svol[dam_oid - 1] += dam[dor.fd.STOR_MCM]

                    new_dor = dor.get_dor(disch[dam_oid - 1], svol[dam_oid - 1])
                    streams[dam_oid - 1][dor_field] = new_dor

                    dw = ndoid[dam_oid - 1]
                    if dw != 0:
                        new_node = dw

                dam_oid = new_node

        return streams

    @staticmethod
    def get_dor(discharge, storage):
        """
        Get the DOR value given the discharge and storage values

        :param discharge: value provided in cubic meters per second
        :param storage: value provided in million cubic meters
        :return: degree of regulation (in percent)
        """
        if discharge == 0:
            return 0

        storage = storage * 1000000  # convert to cubic meters
        # convert cms to annual discharge in cubic meters
        annual_disch_reach = discharge * 60 * 60 * 24 * 365
        temp = storage / annual_disch_reach
        _dor = (100 * temp)

        # Limit DOR to 100. The DOR can technically be higher, however we consider
        # a DOR of 100 % as severely impacting the river and assume this value as
        # the ceiling of the DOR value.
        if _dor > 100:
            return 100

        # A DOR smaller than 0.1 is negligible and should be considered as zero.
        if _dor < 0.1:
            return 0

        return _dor

# define functions from the original .../indices/sed.py module
class sed(object):

    fd = var

    def __init__(self):
        pass

    @staticmethod
    def lakes_calculate(lakes, messages):
        """
        Calculate the trapping of sediments in natural lakes

        :param lakes: numpy array with properties of lakes
        :return: dictionaries of small and large lakes and their trapping in tons per year
        """
        sed.prt("", messages)
        sed.prt("***************************************************", messages)
        sed.prt("PART 1: Calculating sediment trapping in lakes", messages)
        sed.prt("***************************************************", messages)
        sed.prt("", messages)

        # Select natural lake, exclude Grand dam, or regulated lake
        arr1 = lakes[lakes["Lake_type"] == 1]
        # Exclude KCL (GOOD2) dam used in the FFR analysis
        arr2 = arr1[arr1["GOOD"] == 0]
        # Select only lakes that are within stream catchment
        # (some lakes are outside , coastal etc.)
        lakes = arr2[arr2["IN_CATCH"] == 1]

        # Dictionaries will hold values
        small_lake_loss_dict = {}
        large_lake_volu_dict = {}

        # Iterate through the lakes and extract values, fill dictionaries etc.
        for lake in lakes:
            goid = lake["GOID"]

            if lake["IN_STREAM"] == 0:
                # not in stream network, calculate TE_brune and loss
                lake_te = sed.TE(lake["Vol_total"], lake["Dis_avg"])
                lake["TE_brune"] = lake_te

                lake["LOSS_LKES_OUT_NET"] = lake["TE_brune"] * lake["SED_ACC"]

                curr_loss = small_lake_loss_dict.get(goid, 0)
                small_lake_loss_dict[goid] = curr_loss + lake["LOSS_LKES_OUT_NET"]

            else:
                # sum volume_mcm for later processing during routing
                current_lake = large_lake_volu_dict.get(goid, 0)
                large_lake_volu_dict[goid] = current_lake + lake["Vol_total"]

        return small_lake_loss_dict, large_lake_volu_dict

    @staticmethod
    def calculate_sed(streams, dam_volu_dict, lake_volu_dict, small_lake_loss_dict, messages):
        """
        Calculates the Sediment Trapping Index (SED)

        :param streams:
        :param dam_volu_dict:
        :param lake_volu_dict:
        :param small_lake_loss_dict:
        :return:
        """

        # 1) Accumulate sediments taking into account lakes inside and outside
        # the network
        #
        # Calculating potential sediment load

        sed.prt("", messages)
        sed.prt("***************************************************", messages)
        sed.prt("PART 2: Calculating potential sediment load", messages)
        sed.prt("***************************************************", messages)
        sed.prt("", messages)

        for stream in streams:

            dis = stream[sed.fd.DIS_AV_CMS]
            noid = stream[sed.fd.NOID]

            vol_lakes_in_network = lake_volu_dict.get(noid, 0)

            loss_lakes_outside_net = small_lake_loss_dict.get(noid, 0)

            sed_sum_current = stream[sed.fd.ERO_YLD_TON]
            sed_nat_ups = stream[sed.fd.SED_NAT_UP]

            sed_nat = sed_sum_current + sed_nat_ups

            # 1) Account for losses from natural lakes outside network (Type 1)
            sed_nat = sed_nat - loss_lakes_outside_net

            # Write into table
            stream[sed.fd.SED_LSS_LKS_OT_NAT] = loss_lakes_outside_net

            # 2) Account for losses from natural lakes inside network (Type 1)
            loss_lakes_in_nat = sed_nat - (
                    sed_nat * sed.TE(vol_lakes_in_network, dis))

            # Write into table
            stream[sed.fd.SED_LSS_LKS_IN_NAT] = loss_lakes_in_nat

            sed_nat = sed_nat - loss_lakes_in_nat

            # 3) Write to table
            stream[sed.fd.SED_NAT] = sed_nat

            # Add the results to next downstream reach
            if streams[stream[sed.fd.NOID] - 1][sed.fd.NDOID] != 0:
                streams[stream[sed.fd.NDOID] - 1][sed.fd.SED_NAT_UP] = \
                    streams[stream[sed.fd.NDOID] - 1][sed.fd.SED_NAT_UP] + sed_nat

        # 2) Accumulate sediments taking into account lakes inside and
        # outside the network and dams

        # Calculating anthropogenic sediment load

        sed.prt("", messages)
        sed.prt("***************************************************", messages)
        sed.prt("PART 3: Calculating anthropogenic sediment load    ", messages)
        sed.prt("***************************************************", messages)
        sed.prt("", messages)

        for stream in streams:

            dis = stream[sed.fd.DIS_AV_CMS]
            noid = stream[sed.fd.NOID]

            vol_lakes_in_network = lake_volu_dict.get(noid, 0)
            vol_dams = dam_volu_dict.get(noid, 0)

            loss_lakes_outside_net = small_lake_loss_dict.get(noid, 0)

            sed_sum_current = stream[sed.fd.ERO_YLD_TON]
            sed_ant_ups = stream[sed.fd.SED_ANT_UP]

            sed_ant = sed_sum_current + sed_ant_ups

            # 1) Account for losses from natural lakes outside network (Type 1)
            sed_ant = sed_ant - loss_lakes_outside_net

            # Write into table
            stream[sed.fd.SED_LSS_LKS_OT_ANT] = loss_lakes_outside_net

            # 2) Account for losses from natural lakes inside network
            # (Type 1) and dams

            # First process the lakes, ....
            loss_lakes_in_ant = sed_ant - (
                    sed_ant * sed.TE(vol_lakes_in_network, dis))
            # ... and substract the losses from lakes, ....
            sed_ant = sed_ant - loss_lakes_in_ant
            # then process the dams.....
            loss_dams_in_ant = sed_ant - (
                    sed_ant * sed.TE(vol_dams, dis))
            # ... and substract the losses from dams
            sed_ant = sed_ant - loss_dams_in_ant

            # Write into table
            stream[sed.fd.SED_LSS_LKS_IN_ANT] = loss_lakes_in_ant

            # Write into table
            stream[sed.fd.SED_LSS_DMS_ANT] = loss_dams_in_ant

            # 3) Write to table
            stream[sed.fd.SED_ANT] = sed_ant

            # Add the results to next downstream reach
            if streams[stream[sed.fd.NOID] - 1][sed.fd.NDOID] != 0:
                streams[stream[sed.fd.NDOID] - 1][sed.fd.SED_ANT_UP] = \
                    streams[stream[sed.fd.NDOID] - 1][sed.fd.SED_ANT_UP] + sed_ant

        # 3) Calculate the difference between 3 and 2, which is the
        # losses due to dams
        #
        # Calculate index (difference)

        sed.prt("", messages)
        sed.prt("***************************************************", messages)
        sed.prt("PART 4: Calculating SED index", messages)
        sed.prt("***************************************************", messages)
        sed.prt("", messages)

        for stream in streams:
            sed_nat = stream[sed.fd.SED_NAT]
            sed_ant = stream[sed.fd.SED_ANT]

            sediment_loss = sed_nat - sed_ant

            stream[sed.fd.SED_LSS_TOT] = sediment_loss

            # Make sure there is no division by zero error
            if sed_nat > 0.000000001:

                sti = 100 * (sediment_loss / sed_nat)

                # Clip values smaller than 0.1 just like for DOR
                if sti >= 0.1:
                    stream[sed.fd.SED] = 100 * (sediment_loss / sed_nat)
                else:
                    stream[sed.fd.SED] = 0
            else:
                stream[sed.fd.SED] = 0

        return streams

    @staticmethod
    def TE(volume, discharge):
        """
        Calculates the trapping efficiency of reservoir or lake according to the Brune equation

        :param volume: volume of reservoir or lake in million cubic meters (MCM)
        :param discharge: discharge at reservoir or lake location (outflow) in cubic meters per
        second (CMS)

        :return: trapping efficiency in percent
        """

        if discharge < 0.00000001:
            TE_RATIO = 1.0
        else:
            vol_dis = (volume * 1000000.0) / (
                    discharge * 60.0 * 60.0 * 24.0 * 365.0)
            div = math.sqrt(vol_dis)

            if div < 0.00000001:
                TEF = 0
                TE_RATIO = 1.0 - TEF
            else:
                TEF = 1.0 - ((0.05) / float(div))

                if TEF < 0:
                    TEF = 0

                TE_RATIO = 1.0 - TEF

        return TE_RATIO

    @staticmethod
    def prt(txt, messages):
        logging.info(txt)
        messages.AddMessage(txt)

# define the functions from the original .../indices/sta.py module
class sta(object):

    fd = var

    def __init__(self):
        pass

    @staticmethod
    def calculate_sta(stream_array, stream_alt, ff_field, ffr_stat1_field, ffr_stat2_field,
                    ffr_dis_id_field, messages):
        """
        Calculate the free-flowing status (STA)

        :param stream_array:
        :param stream_alt:
        :param ff_field: field name of field indicating CSI above/below threshold
        :param ffr_stat1_field: field name for FFR status (two categories)
        :param ffr_stat2_field: field name for FFR status (three categories)
        :param ffr_dis_id_field: field name for river stretch.
        :return:
        """
        status_dict = {}

        # creating a DataFrame using the alternative (filtered) stream network
        df = pd.DataFrame(stream_alt)

        fun = {sta.fd.LENGTH_KM: np.sum, sta.fd.BB_LEN_KM: 'first'}
        df2 = df.groupby([sta.fd.BB_ID, ffr_dis_id_field, ff_field],
                        as_index=False).agg(fun)

        # TODO: If study area cuts part of river network, the status attribute
        #  will not be correct because the some rivers are cut, and even if all
        #  reaches are FF, will not reach 100 % compared to BB_LEN_KM,
        #  which relates to entire river. Solution: recalculate the BB_LEN_KM
        #  attribute here

        # Calculating percent free-flowing of each stretch using the alternative stream array
        df2.loc[:, "PCT_FF"] = (df2[sta.fd.LENGTH_KM] / df2[sta.fd.BB_LEN_KM]) * 100

        tools.prt(messages,"Creating status dictionaries")
        dis_id = df2[ffr_dis_id_field].tolist()
        pct_ff = df2["PCT_FF"].tolist()
        status = df2[ff_field].tolist()

        x = 0
        for r in dis_id:
            dis_id_val = dis_id[x]
            pct_ff_val = pct_ff[x]
            status_val = status[x]

            status_dict[dis_id_val] = sta.return_sta(status_val=status_val,
                                                pct_ff_val=pct_ff_val)

            x += 1

        tools.prt(messages,"Updating status in original stream_array")
        stream_array = sta.calc_status_values(stream_array, status_dict, ffr_dis_id_field, ffr_stat1_field, ffr_stat2_field, messages)

        return stream_array

    @staticmethod
    def calc_status_values(stream_array, status_dict, ffr_dis_id_field, ffr_stat1_field, ffr_stat2_field, messages):
        """
        Assigns Free-flowing river status (STA) to the river reach.
        There are two types of STA values:

        a) CSI_FF1:

        1 = Free-flowing river; or
        3 = Not a free-flowing river

        b) CSI_FF2:

        1 = Free-flowing river;
        2 = River stretch with "Good Connectivity Status; or
        3 = Not a free-flowing river

        :param stream_array: original stream array to be updated
        :param status_dict: dictionary of FFR status values from alternative stream network
            to be transferred to original stream network
        :param ffr_dis_id_field: river stretch identifier
        :param ffr_stat1_field: field name of FFR status (two categories)
        :param ffr_stat2_field: field name of FFR status (three categories)
        :return:
        """

        for stream in stream_array:
            dis_id = stream[ffr_dis_id_field]
            status = status_dict.get(dis_id, 0)

            if status == 3:
                stream[ffr_stat1_field] = 3
                stream[ffr_stat2_field] = 3
            elif status == 2:
                stream[ffr_stat1_field] = 3
                stream[ffr_stat2_field] = 2
            elif status == 1:
                stream[ffr_stat1_field] = 1
                stream[ffr_stat2_field] = 1
            else:
                # Should not occur
                tools.prt(messages,"{} with invalid STA value".format(str(stream[sta.fd.GOID])))

                stream[ffr_stat1_field] = 0
                stream[ffr_stat2_field] = 0

        return stream_array

    @staticmethod
    def return_sta(status_val, pct_ff_val):
        """
        Calculates the STA value

        :param status_val: CSI status value (0 or 1 depending on threshold)
        :param pct_ff_val: percentage length of river that is free-flowing
        :return: STA value (1, 2, or 3)
        """

        # Making sure percentage cannot be None
        if np.isnan(pct_ff_val):
            pct_ff_val = 0

        if status_val == 1.0:
            if pct_ff_val >= 99.999:
                return 1
            elif pct_ff_val < 99.999:
                return 2
        elif status_val == 0:
            return 3

    @staticmethod
    def dissolve_rivers(stream_array_temp, ff_fields, dis_id_field):
        """
        This function is dissolving the results feature class and return
        aggregated results for each backbone river.

        :param stream_array_temp:  A Feature Class created by joining the CSI results table
                to a simple stream FC as preparation for dissolving
        :param ff_fields:  This field determines if a reach is below (0) or above (1)
                the csi threshold
        :param dis_id_field: Returns dissolved and aggregated FC with statistics to feed
                into filtering (next step)
        :return:
        """

        stream_array_temp[dis_id_field] = 0

        diss_field1 = str(sta.fd.BB_ID)
        diss_field2 = str(ff_fields)

        diss1 = stream_array_temp[diss_field1].tolist()
        diss2 = stream_array_temp[diss_field2].tolist()

        dissid = stream_array_temp[dis_id_field].tolist()

        upsids = stream_array_temp["NUOID"].tolist()

        nodes = []
        diss_id_reach = 0

        # We need a set to check which ids are already in use. IDs should be unique
        diss_id_set = set([])

        sink_streams = stream_array_temp[stream_array_temp["NDOID"] == 0]
        sink_streams = sink_streams["NOID"].tolist()

        x = 0
        for sink in sink_streams:
            x += 0
            """ Process each sink stream in a loop """
            # Make sure the new id has not been used yet
            while True:
                new_id = diss_id_reach + 1
                if new_id not in diss_id_set:
                    diss_id_set.add(new_id)
                    break

            nodes.append(sink)
            dissid[sink - 1] = new_id
            diss_id_reach = new_id

            while True:
                new_nodes = []
                if len(nodes) == 0:
                    break
                for node in nodes:
                    if node != -1 and node != '':

                        # update segment
                        diss_value1 = diss1[node - 1]
                        diss_value2 = diss2[node - 1]

                        diss_id = dissid[node - 1]

                        ups = upsids[node - 1].split("_")

                        if len(ups) >= 1:
                            for up in ups:
                                if up != '':
                                    new_nodes.append(int(up))
                                    # Convert to int and adjust for index start at
                                    # 0
                                    up = int(up) - 1
                                    if (diss1[up] == diss_value1) & (
                                            diss2[up] == diss_value2):
                                        dissid[up] = diss_id
                                    else:
                                        while True:
                                            new_id = diss_id_reach + 1
                                            if new_id not in diss_id_set:
                                                diss_id_set.add(new_id)
                                                break
                                        dissid[up] = new_id
                                        diss_id_reach = new_id
                nodes = new_nodes

        # Assign list to numpy array field
        stream_array_temp[dis_id_field] = dissid

        return stream_array_temp

    @staticmethod
    def apply_volume_filter(csi_fc, ff_field, dis_id_field, pct_aff_thres, messages):

        """
        This filter averages out some small inconsistencies, that could affect big rivers

        :param csi_fc: Path to feature class to filter
        :param ff_field: This field determines if a reach is below (0) or above (1) the csi threshold
        :param dis_id_field: Filtering on (1) or off (0)
        :param pct_aff_thres: volume_mcm threshold
        :return:
        """



        df = pd.DataFrame(csi_fc)
 
        df = df[df[ff_field] == 0]

        if len(df) == 0:
            arcpy.AddError("No river reaches were below the CSI threshold.")
            sys.exit()
        
        fun = {sta.fd.LENGTH_KM: np.sum, sta.fd.VOLUME_TCM: np.sum, sta.fd.BB_VOL_TCM: 'first'}

        df1 = df.groupby([sta.fd.BB_ID, dis_id_field], as_index=False).agg(fun)

        # Calculating a "Length Category"; grouping the backbone rivers into three
        # categories
        df1.loc[:, "PCT_AFF"] = (df1[sta.fd.VOLUME_TCM] / df1[sta.fd.BB_VOL_TCM]) * 100
        df1.loc[:, "FILTER"] = 0

        df1 = df1[df1["PCT_AFF"] < pct_aff_thres]
        df1.loc[:, "FILTER"] = 1

        return df1

    @staticmethod
    def update_csi(streams_diss, bb_id_to_filter, dis_id_field, csi_field, ff_field, messages):
        """
        This function alters the reach level results according to the spatial
        filtering after dissolving that occurred before.

        :param streams_diss: The river reach array to be modified
        :param disk_csi_layer: The river reach fc used for to select from
        during spatial selection
        :param diss_fc: the dissolved backbone layer used for spatial selection
        :param csi_field: field to be altered
        :param ff_field: field to be altered
        :return: modifies "df_mod"
        """

        # The CSI results will be temporarily overwritten, so that the
        # ffr status can be calculated using the previously applied filtering
        # techniques. Original CSI output is not affected
        tools.prt(messages,"Temporary overwriting CSI results")
        cnt = 0

        bbid_list = []
        for index, stretch in bb_id_to_filter.iterrows():
            bbid = int(stretch[sta.fd.BB_ID])
            dis_id = int(stretch[dis_id_field])
            bbid_list.append(str(bbid) + str(dis_id))

        for stream in streams_diss:
            bb_id = int(stream[sta.fd.BB_ID])
            dis_id = int(stream[dis_id_field])

            combo = str(bb_id) + str(dis_id)

            if combo in bbid_list:
                stream[csi_field] = 100
                stream[ff_field] = 1
                cnt += 1

        # Number of river reaches temporarliy altered for this procedur
        tools.prt(messages,"Number of river reaches altered: " + str(cnt))

        return streams_diss

    @staticmethod
    def update_streams_with_diss_id(stream_array, streams_diss2, ff_dis_id_field):
        """
        Function to update original stream ID

        :param stream_array:
        :param streams_diss2:
        :param ff_dis_id_field:
        :return:
        """
        dis_ids = streams_diss2[ff_dis_id_field].tolist()

        indx = 0
        for stream in stream_array:
            disid = dis_ids[indx]
            stream[ff_dis_id_field] = disid
            indx += 1
        return stream_array


########## start functions ##########
class fra_start(object):

    fd = var

    def __init__(self):
        pass

    @staticmethod
    def start(stamp, sequence, para, scenarios, st_flds, paths, messages):
        """
        Starts the assessment and executes the scripts indicated in the Excel file

        :param stamp: time stamp
        :param sequence: list of scripts and weather they should run
        :param para: dictionary of parameters from Excel file
        :param scenarios: List of scenarios from Excel file
        :param st_flds: required fields in stream database
        :param paths: dictionary of paths as defined by setup function
        :return:
        """

        if sequence == ["NO", "NO", "NO", "NO"]:
            tools.prt(messages,"Nothing to run. Exit!")
            sys.exit(0)
        else:
            pass

        if sequence["run_dof"] == "YES":
            fra_start.prt('\n' + "*********", messages)
            fra_start.prt("RUN DOF", messages)
            fra_start.prt("*********" + '\n', messages)

            ffr_run_dof.run_dof(stamp, para, paths, messages)
        if sequence["run_dor"] == "YES":
            fra_start.prt('\n' + "*********", messages)
            fra_start.prt("RUN DOR", messages)
            fra_start.prt("*********" + '\n', messages)

            ffr_run_dor.run_dor(stamp, para, paths, messages)

        if sequence["run_sed"] == "YES":
            fra_start.prt('\n' + "*********", messages)
            fra_start.prt("RUN SED", messages)
            fra_start.prt("*********" + '\n', messages)

            ffr_run_sed.run_sed(para, paths, messages)

        if sequence["run_csi"] == "YES":
            fra_start.prt('\n' + "*********", messages)
            fra_start.prt("RUN CSI", messages)
            fra_start.prt("*********" + '\n', messages)

            ffr_run_csi.run_csi(stamp, para, scenarios, st_flds, paths, messages)

    @staticmethod
    def setup(base, out, xls_full, stamp, xls_file_name, messages):
        """
        Setup assessment

        :param base: local path of assessment from where the script is launched
        :param out: output folder
        :param xls_full: full path and file name of config.xls file
        :param stamp: time stamp
        :param xls_file_name: file name of excel file
        :return:
        """
        # Create CSI geodatabase
        gdb_name = "CSI"
        gdb_full_path, gdb_file_name = tools.create_gdb(out, gdb_name)

        # Make statistics folder, where excel files are stored
        sta_folder = os.path.join(out, "STAT")
        tools.create_path(sta_folder, messages)

        # Make temp csi global_stats.py folder
        sta_csi_folder = os.path.join(out, "STATS_CSI")
        tools.create_path(sta_csi_folder, messages)

        # Make temp csi global_stats.py folder
        sta_pickle_folder = os.path.join(out, "STATS_PICKLES")
        tools.create_path(sta_pickle_folder, messages)

        # Make test_data pickle folder
        test_pickle_folder = os.path.join(out, "TEST_PICKLES")
        tools.create_path(test_pickle_folder, messages)

        # Copy Excel-file into results folder for documentation
        #shutil.copy(xls_full, out + "\\" + xls_file_name)

        # Copy entire code directory into output folder for reference
        #cde_folder = os.path.join(out, "CODE")
        #tools.create_path(cde_folder, messages)
        #tools.copytree(src=base, dst=cde_folder, messages = messages)

        # Copy Python code into results folder for documentation
        #python_code = os.path.abspath(__file__)
        #tail = os.path.basename(python_code)
        #shutil.copy(xls_full, out + "\\" + tail)

        # Setup logging
        tools.setup_logging(out, messages)

        # Create Excel writer
        writer, excel_file = tools.get_writer(sta_folder, stamp)
        tools.create_results_sheet(writer)

        paths = {"writer": writer,
                "excel_file": excel_file,
                "gdb_full_path": gdb_full_path,
                "sta_csi_folder": sta_csi_folder,
                "sta_pickle_folder": sta_pickle_folder,
                "test_pickle_folder": test_pickle_folder}
                


        return paths


    @staticmethod
    def prt(txt, messages):
        logging.info(txt)
        messages.AddMessage(txt)

    @staticmethod
    def main(messages):
        
        messages.addMessage("Starting model at {}".format(datetime.datetime.now()))

        path = os.path.realpath(__file__)
        basepath, filename = os.path.split(path)

        xls_full_path = var.config_excel

        xls_path, xls_filename = os.path.split(xls_full_path)

        # Loading model parameters from EXCEL file
        sequence, para, scenarios, fields = tools.load_parameters(xls_full_path, messages)
        
        # These fields must be present in source dataset feature class
        st_flds = [fra_start.fd.REACH_ID, fra_start.fd.GOID, fra_start.fd.NOID, fra_start.fd.NUOID, fra_start.fd.NDOID,
                fra_start.fd.CON_ID, fra_start.fd.BAS_ID, fra_start.fd.BAS_NAME, fra_start.fd.RIV_ORD,
                fra_start.fd.DIS_AV_CMS, fra_start.fd.LENGTH_KM,
                fra_start.fd.VOLUME_TCM, fra_start.fd.ERO_YLD_TON, fra_start.fd.HYFALL, fra_start.fd.BB_ID, fra_start.fd.BB_NAME,
                fra_start.fd.BB_OCEAN, fra_start.fd.ISO_NAME,
                fra_start.fd.BB_LEN_KM, fra_start.fd.BB_VOL_TCM, fra_start.fd.BB_DIS_ORD, fra_start.fd.INC
                ] + fields + [fra_start.fd.FLD]

        ba_flds = [para["svol_field"]]

        tools.check_fields(para["streams_fc"], st_flds, messages)
        tools.check_fields(para["dams_fc"], ba_flds, messages)

        # 1 Setting up file structure
        paths = fra_start.setup(base=basepath,
                    out=var.output_folder,
                    xls_full=xls_full_path,
                    stamp=var.time_stamp,
                    xls_file_name=xls_filename, 
                    messages = messages)

        fra_start.prt("Results will be in: " + str(var.output_folder), messages)

        fra_start.start(var.time_stamp, sequence, para, scenarios, st_flds, paths, messages)


########## create the tool ##########
class Tool(object):
    
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Free_Flowing_Rivers_Analysis"
        self.description = "Free_Flowing_Rivers_Analysis"
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""

        ##### START sheet parameters #####
        # the settings sheet scenarios sheet names will be hard coded
        # the only parameters needed from the start page are the run options 
    
        # first parameter: run options as multivariate buttons
        param0 = arcpy.Parameter(
            displayName='Run options', 
            name='run_options', 
            datatype='GPString', 
            parameterType='Required', 
            direction="Input", 
            multiValue=True,
            category = "1. Run Options")            

        
        # applies a filter to the parameter to limit user choices
        param0.filter.type = "ValueList"

        # provides the names of the options availble to the user
        param0.filter.list = ["DOF", "DOR", "SED", "CSI"]
        

        ##### Settings sheet parameters #####
        
        ### path settings
        # 5 path settings are required, dams_fc, bench_fc, streams_fc, lakes_fc, output_folder
        
        # parameter for the path to the dam feature class
        param1 = arcpy.Parameter(
            displayName='Barrier feature class', 
            name='Barrier_feature_class', 
            datatype='GPFeatureLayer', 
            parameterType='Required', 
            direction="Input", 
            multiValue= False,
            category = "2. Path Settings") 

        # parameter for the path to the benchmark feature class
        param2 = arcpy.Parameter(
            displayName='Benchmark feature class', 
            name='benchmark_feature_class', 
            datatype='GPFeatureLayer', 
            parameterType='Required', 
            direction="Input", 
            multiValue= False,
            category = "2. Path Settings") 

        # parameter for the path to the stream feature class
        param3 = arcpy.Parameter(
            displayName='Streams feature class', 
            name='streams_feature_class', 
            datatype='GPFeatureLayer', 
            parameterType='Required', 
            direction="Input", 
            multiValue= False,
            category = "2. Path Settings") 

        # parameter for the path to the lake feature class
        param4 = arcpy.Parameter(
            displayName='Lakes feature class', 
            name='lakes_feature_class', 
            datatype='GPFeatureLayer', 
            parameterType='Required', 
            direction="Input", 
            multiValue= False,
            category = "2. Path Settings") 

        # parameter for the path to the output folder
        param5 = arcpy.Parameter(
            displayName='Output folder', 
            name='output_folder', 
            datatype='DEFolder', 
            parameterType='Required', 
            direction="Input", 
            multiValue= False,
            category = "2. Path Settings") 


        ### general settings
        # 3 general settings are available, barrier_inc_field, update_mode, min_length
        
        # parameter for the field in the dams/barrier database that determines which dams are used in the assessment
        param6 = arcpy.Parameter(
            displayName='Dam/Barrier Inclusion Field', 
            name='barrier_inclusion_field', 
            datatype='Field', 
            parameterType='Required', 
            direction="Input", 
            multiValue= False,
            category = "3. General Settings" )

        # the options in the field are dependent on the fields available in the dam feature class
        # set the dependency to the fields of the dam feature class
        param6.filter.list = []                     # think about if the field type should be filtered to limit invalid field options
        param6.parameterDependencies = [param1.name]

        # parameter to specify if results should be written drectly into the streams table (overwrite). YES or NO
        # drop down would be nice here.
        param7 = arcpy.Parameter(
            displayName='Update/Overwrite stream table?', 
            name='update_mode', 
            datatype='GPString', 
            parameterType='Required',  
            direction="Input", 
            multiValue= False,
            category = "3. General Settings" )       

        # applies a filter to the parameter to limit user choices
        param7.filter.type = "ValueList" 
        # provides the options availble to the user
        param7.filter.list = ["YES", "NO"]

        # parameter for the length threshold for statisics generating 1) a list of free flowing rivers and 2) a list of good connectivity segments     
        param8 = arcpy.Parameter(
            displayName='Minimum Stream Length Threshold', 
            name='min_length', 
            datatype='GPDouble', 
            parameterType='Required',  
            direction="Input", 
            multiValue= False,
            category = "3. General Settings")


        ### DOF settings ###
        # 8 DOF settings are available: dof_field, drf_upstream, drf_downstream, use_dam_level_df, dof_mode, dof_dis_mode, str_lvl_field, bar_dis_field

        # parameter for the field to use when updating DOF
        param9 = arcpy.Parameter(
            displayName='Degree of Fragmentation field', 
            name='DOF_field', 
            datatype='Field', 
            parameterType='Required',  
            direction="Input", 
            multiValue= False,
            category = "4. Degree of Fragmentation Settings")

        # the options for fields is dependent on the fields available in the stream feature class
        # set the dependency to the fields of the stream feature class
        param9.filter.list = []                     # think about if the field type should be filtered to limit invalid field options
        param9.parameterDependencies = [param3.name]

        # parameter for use_dam_level_df
        param10 = arcpy.Parameter(
            displayName='Use dam level discharge factors?', 
            name='use_dam_level_df', 
            datatype= 'GPString', 
            parameterType='Required',  
            direction="Input", 
            multiValue= False,
            category = "4. Degree of Fragmentation Settings")
        
        # applies a filter to the parameter to limit user choices
        param10.filter.type = "ValueList" 
        # provides the options availble to the user
        param10.filter.list = ["YES", "NO"]

        # parameter for discharge range factor used in DOF calculations for upstream direction
        param11 = arcpy.Parameter(
            displayName='Upstream Discharge Range Factor', 
            name='drf_upstream', 
            datatype='GPLong', 
            parameterType='Optional', 
            direction="Input", 
            multiValue= False,
            enabled = False,
            category = "4. Degree of Fragmentation Settings")

        # parameter for discharge range factor used in DOF calculations for downstream direction
        param12 = arcpy.Parameter(
            displayName='Downstream Discharge Range Factor', 
            name='drf_downstream', 
            datatype='GPLong', 
            parameterType='Optional',  
            direction="Input", 
            multiValue= False,
            enabled = False,
            category = "4. Degree of Fragmentation Settings")

        # parameter for upstream discahrge factor field 
        param13 = arcpy.Parameter(
            displayName='Dam Level Upstream Discharge Factor Field', 
            name='DFU_field', 
            datatype= 'field', 
            parameterType='Optional',  
            direction="Input", 
            multiValue= False,
            enabled = False,
            category = "4. Degree of Fragmentation Settings")               

        # the options in the field are dependent on the fields available in the dam feature class
        # set the dependency to the fields of the dam feature class
        param13.filter.list = []                     # think about if the field type should be filtered to limit invalid field options
        param13.parameterDependencies = [param1.name]

        # parameter for downstream discharge factor field
        param14 = arcpy.Parameter(
            displayName='Dam Level Downstream Discharge Factor Field', 
            name='DFD_field', 
            datatype= 'field', 
            parameterType='Optional', 
            direction="Input", 
            multiValue= False,
            enabled = False,
            category = "4. Degree of Fragmentation Settings")               

        # the options in the field are dependent on the fields available in the dam feature class
        # set the dependency to the fields of the dam feature class
        param14.filter.list = []                     # think about if the field type should be filtered to limit invalid field options
        param14.parameterDependencies = [param1.name]



        # parameter for DOF decay function type.
        param15 = arcpy.Parameter(
            displayName='Degree of Fragmentation Decay Function Type ', 
            name='dof_mode', 
            datatype= 'GPLong', 
            parameterType='Derived', 
            multiValue= False,
            category = "4. Degree of Fragmentation Settings")               

        # hard code the DOF decay function value as 1
        param15.value = 1

        ### DOR settings
        # 2 DOR settings are available: dor_field, svol_field

        # parameter for the field to use for updating Degree of regulation
        param16 = arcpy.Parameter(
            displayName='Field for Updating Degree of Regulation', 
            name='dor_field', 
            datatype= 'Field', 
            parameterType='Required', 
            direction="Input", 
            multiValue= False,
            category = "5. Degree of Regulation Settings") 

        # the options for fields is dependent on the fields available in the stream feature class
        # set the dependency to the fields of the stream feature class
        param16.filter.list = []                     # think about if the field type should be filtered to limit invalid field options
        param16.parameterDependencies = [param3.name]

        # parameter for the the source field for reservoir strage volume
        param17 = arcpy.Parameter(
            displayName=' Source Field for Reservoir Storage Volume', 
            name='svol_field', 
            datatype= 'Field', 
            parameterType='Required',  
            direction="Input", 
            multiValue= False,
            category = "5. Degree of Regulation Settings") 

        # the options for fields is dependent on the fields available in the stream feature class
        # set the dependency to the fields of the stream feature class
        param17.filter.list = []                     # think about if the field type should be filtered to limit invalid field options
        param17.parameterDependencies = [param1.name]

        ### SED settings
        # 1 SED setting is available: sed_field

        # parameter for the source field for sediment trapping index to be used in CSI calculations
        param18 = arcpy.Parameter(
            displayName=' Source Field for Sediment Trapping Index', 
            name='sed_field', 
            datatype= 'Field', 
            parameterType='Required', 
            direction="Input", 
            multiValue= False,
            category = "6. Sediment Trapping Index Settings") 

        # the options for fields is dependent on the fields available in the stream feature class
        # set the dependency to the fields of the stream feature class
        param18.filter.list = []                     # think about if the field type should be filtered to limit invalid field options
        param18.parameterDependencies = [param3.name]        

        
        
        ##### Scenario sheet parameters #####

        param19 = arcpy.Parameter(
        displayName='Scenario Name',
        name='scenario_settings',
        datatype='GPValueTable',
        parameterType='Required',
        direction='Input',
        category = "7. Scenario Settings")

        # provides the datatype and columns name for the value table
        param19.columns = [
            ['GPString', 'Scenario name'],
            ['GPString', "dof_indicator"],
            ['GPString', "dor_indicator"],
            ['GPString', "sed_indicator"],
            ['GPString', "use_indicator"],
            ['GPString', "rdd_indicator"],
            ['GPString', "urb_indicator"],
            ['GPDouble', "dof_weight"],
            ['GPDouble', "dor_weight"],
            ['GPDouble', "sed_weight"],
            ['GPDouble', "use_weight"],
            ['GPDouble', "rdd_weight"],
            ['GPDouble', "urb_weight"],
            ['GPDouble', 'CSI_threshold'],
            ['GPDouble', 'fld_damp'],
            ['GPDouble', 'filter_threshold'],
            ['GPString', 'to_process'],
            ['GPString', 'to_export']    
            ]
        
        # provides the options availble to the user for to_process and to_export
        param19.filters[16].list = ['YES', 'NO']
        param19.filters[17].list = ['YES', 'NO']
        
        # create parameter to autopopulate parameters using an excel sheet
        param20 = arcpy.Parameter(
            displayName='Auto-populate parameters using config excel sheet',
            name='config_excel',
            datatype='File',
            parameterType='Optional',
            direction='Input')

        # restrict file options to .xls or .xlsx
        param20.filter.list = ['xls', 'xlsx']

        # create a boolean derived parameter to track whether an excel sheet was used to 
        # autopopulate the parameters. This will be used to trigger the creation of a message in 
        # update messages if an excel sheet was used to create auto-populate the parameters
        param21 = arcpy.Parameter(
            displayName='generate auto-populate errors',
            name='auto_pop_errors',
            datatype='GPString',
            parameterType='Derived',
            direction='Output')
        

        # create a button that enables users to autopopulate the parameters with default settings
        param22 = arcpy.Parameter(
            displayName = 'Load default settings',
            name='load_default_button',
            datatype='GPBoolean',
            parameterType='Optional',
            direction='Input')
        
        # create a derived variable to track whether the default settings have been loaded
        # this is required to allow users to change the default settings
        # without this, whenever the default settings button is checked, if the user changes a parameter
        # the parameter would be reset to the default value
        param23 = arcpy.Parameter(
            displayName = 'default loaded',
            name='default_loaded',
            datatype='GPBoolean',
            parameterType='Derived',
            direction='Output')
        # assign the value to start as false   
        param23.value = False

        ### put all parameters into a list
        params = [param0, param1, param2, param3, param4, param5,
        param6, param7, param8, param9, param10, param11, param12, param13, param14,
        param15, param16, param17, param18, param19, param20, param21, param22, param23]
        
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def initializeParameters(self, parameters):
        """Refine the properties of a tools parameters. This method is called when the tool is opened."""
        
        
        return

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        

        # create readable variable names
        run_options = parameters[0]
        dams_fc = parameters[1]
        bench_fc = parameters[2]
        streams_fc = parameters[3]
        lakes_fc = parameters[4]
        output_folder_para = parameters[5]
        barrier_inc_field = parameters[6]
        update_mode = parameters[7]
        min_length = parameters[8] 
        dof_field = parameters[9]
        use_dam_level_df = parameters[10]
        drf_upstream = parameters[11] 
        drf_downstream = parameters[12] 
        DFU_field = parameters[13]
        DFD_field = parameters[14]
        dof_mode = parameters[15] 
        dor_field = parameters[16]
        svol_field = parameters[17]
        sed_field = parameters[18]
        scenarios_table = parameters[19]
        config_excel_param = parameters[20]
        autopop_errors_param = parameters[21]
        load_default_button = parameters[22]
        default_loaded = parameters[23]
        


        ##### load defualt settings if selected #####
        # if the button is checked and the parameters have not been loaded
        if load_default_button.value == True and default_loaded.value == False:
            # assign the deafult settings to all of the parameters

            # assign run options
            run_options.value = ["DOF", "DOR", "SED", "CSI"]

            # assign path settings
            # Get path to toolbox
            path = os.path.realpath(__file__)
            # split toolbox path into basepath and filename
            basepath, filename = os.path.split(path)

            # join base_path and data in the datafolder
            dams_fc.value = os.path.join(basepath, "data\\example_data.gdb\\barriers")
            bench_fc.value = os.path.join(basepath, "data\\example_data.gdb\\benchmark_rivers")
            streams_fc.value = os.path.join(basepath, "data\\example_data.gdb\\streams")
            lakes_fc.value = os.path.join(basepath, "data\\example_data.gdb\\lakes")
            # join base_path and output folder
            output_folder_para.value = os.path.join(basepath, r"output")

            # set defaults for General settings
            barrier_inc_field.value = "INC1"
            update_mode.value =  "YES"
            min_length.value = 100

            # set defaults for DOF settings 
            dof_field.value = "DOF"
            use_dam_level_df.value = "NO"
            drf_upstream.value = 5 
            drf_downstream.value = 5 
            DFU_field.value = None
            DFD_field.value = None

            # set defaults for DOR settings
            dor_field.value = "DOR"
            svol_field.value = "STOR_MCM"

            # set defaults for SED settings
            sed_field.value = "SED"

            # set defaults for scenarios table
            scenarios_table.value = [ [ 'CSI01', 'DOF', 'DOR', 'SED', 'USE', 'RDD', 'URB', 30, 30, 15, 15, 5, 5, 95, 50, 0.1, "YES", "YES"]]

            default_loaded.value = True

        # check if the default button is unchecked and if the default parameters had previously been loaded 
        if load_default_button.value == False and default_loaded.value == True:
            # reset the default loaded value to false
            # resetting the loaded parameter allows users to reapply if the default settings
            # if they loaded them, changed something, and want to reload the,=m
            default_loaded.value = False
    


        ##### load excel config parameter settings #####
        # create more readable variable name
        config_excel_param = parameters[20]

        # if there is something for the parameter
        if config_excel_param != None:
            
            # if the config excel file parameter has not been validated
            if not config_excel_param.hasBeenValidated:
                
                # get the parameter value
                config_excel_string = config_excel_param.value
                # use describe to describe the geoprocessing object that the .value returns
                config_excel_desc = arcpy.Describe(config_excel_string)
                # get the path to the excel file from the describe object
                config_excel_path = config_excel_desc.catalogPath

                # create an empty list to append error messages to
                errors_list = []
                
                # create two dictionary objects
                start_dict = defaultdict()
                value_dict = defaultdict()
                
                try:
                    # open the excel file using pandas
                    xls = pd.ExcelFile(config_excel_path)
                
                
                except Exception as e:
                    # create error if the excel sheet cannot be opened
                    errors_list.append(". Unable to open the excel file. {} ".format(str(e)))


                if 'xls' in locals():
                    # get the sheet names from the excel file
                    sheet_names = xls.sheet_names

                    # check if the sheet named START is in the list of sheet names
                    if "START" not in sheet_names:
                        # if no start sheet exists, create error
                        errors_list.append(". The sheet 'START' does not exist. ")


                    # check that the excel file has atleast three sheets
                    if len(sheet_names) < 3:
                        errors_list.append(". There must be at least three sheets in the workbook: 'START', a settings sheet, and a scenario sheet. ")

                    else:
                        try:
                            # parse the start sheet
                            settings_sheet = xls.parse("START")

                        except Exception as e:
                            # create error if start sheet cannot be parsed
                            errors_list.append(". Unable to parse the 'START' sheet. {} ".format(str(e)))
                

                        # create two variables to store values from the key column and
                        # value cloumn of the start sheet
                        if 'settings_sheet' in locals():
                            if "Key" in settings_sheet.keys():
                                start_var1 = settings_sheet['Key']
                            else:
                                errors_list.append(". The column 'Key' does not exist in the 'START' sheet. Ensure the variable is spelled correctly. ")
                            if "Value" in settings_sheet.keys():
                                start_var2 = settings_sheet['Value']
                            else:
                                errors_list.append(". The column 'Value' does not exist in the 'START' sheet. Ensure the variable is spelled correctly. ")

                            # create a counter starting at 0
                            ind = 0
                            
                            if 'start_var1' in locals() and 'start_var2' in locals():
                                # for each entry in var 1
                                for fname in start_var1:
                                    # the value associated with fname in start dict is assigned as
                                    # the corresponding value in var2
                                    start_dict[fname] = start_var2[ind]
                                    # increase the counter by 1
                                    ind += 1

                                # iterates throught the list of strings and checks if they are in the
                                # 'keys' of the start dic. If not an error is added to the list of errors
                                for sname in ["settings_sheet", "scenarios_sheet", "run_dof", "run_dor", "run_sed", "run_csi"]:
                                    if sname not in start_dict.keys():
                                        errors_list.append(". Start variable not found: {}. Parameter values for {} could not be populated. Check that the variable name is spelled correctly in the start sheet. ".format(sname, sname))

                                ### Start sheet parameters ###

                                # assigns the settings and scenarios sheet values in the start_dict to variable names
                                if "settings_sheet" in start_dict.keys():
                                    set_sheet = start_dict["settings_sheet"]
                                    
                                    
                                
                                if "scenarios_sheet" in start_dict.keys():
                                    sce_sheet = start_dict["scenarios_sheet"]

                                # create an empty list to append run options to
                                run_options_list = []
                                
                                # checks if the run settings have be set to yes
                                # if the run setting is set to yes the name of the 
                                #run setting is appended to the run_options_list
                                if "run_dof" in start_dict.keys():
                                    if start_dict["run_dof"] == "YES":
                                        run_options_list.append("DOF")
            
                                if "run_dor" in start_dict.keys():
                                    if start_dict["run_dor"] == "YES":
                                        run_options_list.append("DOR")
                                
                                if "run_sed" in start_dict.keys():
                                    if start_dict["run_sed"] == "YES":
                                        run_options_list.append("SED")
                                
                                if "run_csi" in start_dict.keys():
                                    if start_dict["run_csi"] == "YES":
                                        run_options_list.append("CSI")
            

                                # assign run option values to the parameter
                                run_options.value = run_options_list

            
                                ##### settings sheet parameters ######
                                if 'set_sheet' in locals():
                                    try:
                                        # parses the settings sheet
                                        settings_sheet = xls.parse(set_sheet)

                                    except Exception as e:
                                        # creates an error if the settings sheet cannot be parsed
                                        errors_list.append("." + str(e) + "." + " Does sheet {} exist? ".format(set_sheet))
                                        
                                    # create two variables to store values from the key column and
                                    # value cloumn of the settings sheet
                                    if 'settings_sheet' in locals():
                                        if "Key" in settings_sheet.keys():
                                            set_var1 = settings_sheet['Key']
                                        else:
                                            errors_list.append(". The column 'Key' does not exist in the settings sheet. Ensure the variable is spelled correctly. ")
                                        if "Value" in settings_sheet.keys():
                                            set_var2 = settings_sheet['Value']
                                        else:
                                            errors_list.append(". The column 'Value' does not exist in the settings sheet. Ensure the variable is spelled correctly. ")

                                        # create a counter startng at 0
                                        ind = 0

                                        if 'set_var1' in locals() and 'set_var2' in locals():
                                        # for each entry in var1
                                            for fname in set_var1:
                                                # the value associated with fname in value_dict is assigned the
                                                # corresponding value in var2
                                                value_dict[fname] = set_var2[ind]
                                                # increase the counter by 1
                                                ind += 1

                                            # apply settings values to parameters
                                            # below could be improved with a loop using the parameter.name as the index for value_dict 
                                            # so long as the parameter.name matches the key (might need to modify the parameter names slightly)
                                            
                                            # if the parameter is not in the value_dict an error is created
                                            if 'dams_fc' in value_dict.keys():
                                                if str(value_dict["dams_fc"]) == 'nan':
                                                    dams_fc.value = None
                                                else: 
                                                    dams_fc.value = value_dict["dams_fc"]
                                            else:
                                                errors_list.append(". The Key 'dams_fc' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")

                                            if 'bench_fc' in value_dict.keys():
                                                if str(value_dict["bench_fc"]) == 'nan':
                                                    bench_fc.value = None
                                                else:
                                                    bench_fc.value = value_dict["bench_fc"]
                                            else:
                                                errors_list.append(". The Key 'bench_fc' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'streams_fc' in value_dict.keys():
                                                if str(value_dict["streams_fc"]) == 'nan':
                                                    streams_fc.value = None
                                                else:
                                                    streams_fc.value = value_dict["streams_fc"]
                                            else:
                                                errors_list.append(". The Key 'streams_fc' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                
                                            if 'lakes_fc' in value_dict.keys():
                                                if str(value_dict["lakes_fc"]) == 'nan':
                                                    lakes_fc.value = None
                                                else:
                                                    lakes_fc.value = value_dict["lakes_fc"]
                                            else:
                                                errors_list.append(". The Key 'lakes_fc' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")

                                            if 'output_folder' in value_dict.keys():
                                                if str(value_dict["output_folder"]) == 'nan':
                                                    output_folder_para.value = None
                                                else:   
                                                    output_folder_para.value = value_dict["output_folder"]
                                            else:
                                                errors_list.append(". The Key 'output_folder' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")

                                            if 'barrier_inc_field' in value_dict.keys():
                                                barrier_inc_field.value = value_dict["barrier_inc_field"]
                                            else:
                                                errors_list.append(". The Key 'barrier_inc_filed' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'update_mode' in value_dict.keys():
                                                update_mode.value = value_dict["update_mode"]
                                            else:
                                                errors_list.append(". The Key 'update_mode' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'min_length' in value_dict.keys():
                                                min_length.value = value_dict["min_length"] 
                                            else:
                                                errors_list.append(". The Key 'min_length' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")

                                            if 'dof_field' in value_dict.keys():
                                                dof_field.value = value_dict["dof_field"]
                                            else:
                                                errors_list.append(". The Key 'dof_field' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'use_dam_level_df' in value_dict.keys():
                                                use_dam_level_df.value = value_dict["use_dam_level_df"]
                                            else:
                                                errors_list.append(". The Key 'use_dam_level_fc' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                
                                            # these parameters are optional depending on the setting for use_dam_level_df
                                            # therefore, they may be empty (stored as nan), if they are nan assign the value as none
                                            if 'drf_upstream' in value_dict.keys():
                                                if str(value_dict["drf_upstream"]) == 'nan':
                                                    drf_upstream.value = None
                                                else:        
                                                    drf_upstream.value = value_dict["drf_upstream"] 
                                            else:
                                                errors_list.append(". The Key 'drf_upstream' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'drf_downstream' in value_dict.keys():
                                                if str(value_dict["drf_downstream"]) == 'nan':
                                                    drf_downstream.value = None
                                                else: 
                                                    drf_downstream.value = value_dict["drf_downstream"]
                                            else:
                                                errors_list.append(". The Key 'drf_downstream' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'DFU_field' in value_dict.keys():
                                                if str(value_dict["DFU_field"]) == 'nan':
                                                    DFU_field.value = None
                                                else:
                                                    DFU_field.value = value_dict["DFU_field"]
                                            else:
                                                errors_list.append(". The Key 'DFU_field' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")

                                            if 'DFD_field' in value_dict.keys():
                                                if str(value_dict["DFD_field"]) == 'nan':
                                                    DFD_field.value = None
                                                else:        
                                                    DFD_field.value = value_dict["DFD_field"]
                                            else:
                                                errors_list.append(". The Key 'DFD_field' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")

                                            # apply settings values to parameters
                                            if 'dof_mode' in value_dict.keys():
                                                dof_mode.value = value_dict["dof_mode"] 
                                            else:
                                                errors_list.append(". The Key 'dof_mode' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'dor_field' in value_dict.keys():
                                                dor_field.value = value_dict["dor_field"]
                                            else:
                                                errors_list.append(". The Key 'dor_field' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'svol_field' in value_dict.keys():
                                                svol_field.value = value_dict["svol_field"]
                                            else:
                                                errors_list.append(". The Key 'svol_field' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")
                                            
                                            if 'sed_field' in value_dict.keys():
                                                sed_field.value = value_dict["sed_field"]
                                            else:
                                                errors_list.append(". The Key 'sed_field' does not exist on the settings sheet. No parameter value was applied. Ensure the variable is spelled correctly. ")







                                ### Load scenario list ###
            
                                if 'sce_sheet' in locals():
                                    try:
                                        # parse the scenario sheet
                                        sce_sheet_parsed = xls.parse(sce_sheet)
                                    except Exception as e:
                                        # if the scenario sheet cannot be parsed, create an error
                                        errors_list.append(". " + str(e) + "Does sheet {} exist? ".format(sce_sheet))

                                    if 'sce_sheet' in locals():
                                        
                                        # we index the row using the column name
                                        expected_columns = ['scenario_name', 'dof_indicator', 'dor_indicator', 'sed_indicator', 'use_indicator',
                                                'rdd_indicator', 'urb_indicator', 'dof_weight', 'dor_weight', 'sed_weight', 'use_weight',
                                                'rdd_weight', 'urb_weight', 'csi_thresh', 'fld_damp', 'filter_thresh','to_process','to_export']

                                        # create an empty list of lists that has the same number of empty lists as there are rows in the scenario sheet
                                        # also make the empty lists with empty entries to match the number of expected scenario values to support assigning values by indexing
                                        vtable = [ [None]*len(expected_columns) for list in range(len(sce_sheet_parsed))]
                    

                                        # iterate through each row of the scenarios sheet and append the values to the corresponding list in the vtable
                                        for index, row in sce_sheet_parsed.iterrows():
                                            # iterate through expected column names
                                            for i in range(len(expected_columns)):
                                                # iterate through each column in the row 
                                                for column in row.index:
                                                    # if the row column matches the current expected name being iterated 
                                                    if column == expected_columns[i]:
                                                        # the row value for the column with that name is added to the value table at 
                                                        # the position of the expected column name in its respective list 
                                                        vtable[index][i] = row[expected_columns[i]]
                    

                                            # assign the values of the vtable to the scenarios_table parameter
                                            scenarios_table.values = vtable
                                            
                                            
                # if no error messages were appended to the errors_list above
                if len(errors_list) == 0:
                    # assign the a message to the autopop_errors_param
                    autopop_errors_param.value = ("Parameters have been autopopulated using a config excel file. " + 
                    "Please review the parameters for accuracy.")
                
                # if errors were appended to the list
                else:
                    # create an empty list 
                    errors_string_list = []
                    # count the number of errors in the errors_list
                    num_errors = len(errors_list)
                    # if there is 1 error, create a message
                    if num_errors == 1:

                        errors_string = "Encountered "+ str(num_errors) + " error while loading config excel file. " +"ERROR: " + "1" + "".join(errors_list)
                        errors_string_list.append(errors_string)
                        errors_string_list.append("ALTERNATIVELY, provide the missing parameter values below manually.")
                    # if there are more than 1 error, create a message
                    if num_errors > 1:
                        errors_string_list.append("Encountered "+ str(num_errors) + " errors while loading config excel file. ")
                        for i in range(len(errors_list)):
                            
                        
                            errors_string = "ERROR: " + str(i+1) + "".join(errors_list[i])
                            
                            errors_string_list.append(errors_string)
                        errors_string_list.append("ALTERNATIVELY, provide the missing parameter values below manually.")
                    
                    # join the error string list entries into one string and assign the string to the autopop_errors_param
                    autopop_errors_param.value = "".join(errors_string_list)
                

                # clear the excel config parameter
                config_excel_param.value = None
                
                


        ### create options for Scenario sheet based on fields in the streams_fc ###
        # assign parameter to a more descriptive variable name 
        streams = parameters[3]
        scenario_param = parameters[19]


        if streams.altered:
            if streams.valueAsText != None:
                if arcpy.Exists(streams.valueAsText):

                    # empty list to populate with fields
                    fields_list = []
                            
                    # get parameter
                    fc = streams.valueAsText

                    # check if parameter has been input
                    if fc:

                        # get input feature class fields and add to list
                        fields_list += [f.name for f in arcpy.ListFields(fc)] 

                        # update options available for the indicators in the Scenario table
                        scenario_param.filters[1].list = fields_list   # dof_indicator
                        scenario_param.filters[2].list = fields_list   # dor_indicator
                        scenario_param.filters[3].list = fields_list   # sed_indicator
                        scenario_param.filters[4].list = fields_list   # use_indicator
                        scenario_param.filters[5].list = fields_list   # rdd_indicator
                        scenario_param.filters[6].list = fields_list   # urb_indicator


        ##### change the enabled parameter for parameters affected by use_dam_level_df #####
        # create readable names
        use_dam_level_df = parameters[10]
        drf_upstream = parameters[11] 
        drf_downstream = parameters[12] 
        DFU_field = parameters[13]
        DFD_field = parameters[14]

        # check if the use dam level df has been set to no
        if use_dam_level_df.valueAsText == "NO":
            # enable the static drf value parameters
            drf_upstream.enabled = True
            drf_downstream.enabled = True
            # pre-populate the parameter with a default value of 5 if there is no value
            if drf_upstream.value == None:
                drf_upstream.value = 5
            if drf_downstream.value == None:
                drf_downstream.value = 5
            # disable the field drf value parameters
            DFU_field.enabled = False
            DFD_field.enabled = False
            # clear the disabled parameters
            DFU_field.value = None
            DFD_field.value = None
            
        # check if the use dam level df has been set to yes
        if use_dam_level_df.valueAsText == "YES":
            # enable the field drf value parameters
            DFU_field.enabled = True
            DFD_field.enabled = True
            # disable the static drf value parameters
            drf_upstream.enabled = False
            drf_downstream.enabled = False
            # clear the disabled parameters
            drf_upstream.value = None
            drf_downstream.value = None


        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""


        ##### create message if excel sheet was used to autopopulate parameters #####

        # converter parameters to readable variable names
        config_excel_param = parameters[20]
        autopop_errors_param =  parameters[21]
        
        # if there is a value for the autopop_errors_param, print the message next to the excel parameter. 
        if autopop_errors_param.valueAsText != None:
            config_excel_param.setWarningMessage(autopop_errors_param.valueAsText)
            


        ##### check that required fields are present in the input data #####

        # convert parameters to readable variable names
        barriers = parameters[1] 
        bench_rivs = parameters[2]
        streams = parameters[3]
        lakes = parameters[4]
        
        ### check barriers fc for required fields
        if barriers.altered:
            if barriers.valueAsText != None:
                if arcpy.Exists(barriers.valueAsText):
                    # create list of required field names
                    required_barrier_fields = ["BAS_ID", "GOID", "NOID", "STOR_MCM"]

                    # create a list of fields in barrier feature class
                    barrier_fields = arcpy.ListFields(barriers.valueAsText)
                    # create list of field names from list of fields
                    barrier_field_names = [f.name for f in barrier_fields]

                    # iterate throught the list of required fields and check if it is in barrier feature class
                    for field in required_barrier_fields:
                        if field not in barrier_field_names:
                            barriers.setErrorMessage("The required field '{}' was not found in the barriers feature class. Check that the field exists and is spelled correctly.".format(field)) 
            
        ### check benchmark rivers fc for required fields
        if bench_rivs.altered:
            if bench_rivs.valueAsText != None:
                if arcpy.Exists(bench_rivs.valueAsText):
                    # create list of required field names
                    required_bench_fields = [ "GOID", "BB_ID", "FFRID", "Name_Expert", "BENCH_SRC"]

                    # create a list of fields in barrier feature class
                    bench_fields = arcpy.ListFields(bench_rivs.valueAsText)
                    # create list of field names from list of fields
                    bench_field_names = [f.name for f in bench_fields]

                    # iterate throught the list of required fields and check if it is in barrier feature class
                    for field in required_bench_fields:
                        if field not in bench_field_names:
                            bench_rivs.setErrorMessage("The required field '{}' was not found in the bencmark rivers feature class. Check that the field exists and is spelled correctly.".format(field)) 

        ### check streams fc for required fields
        if streams.altered:
            if streams.valueAsText != None:
                if arcpy.Exists(streams.valueAsText):
                    # create list of required field names
                    required_stream_fields = ["BAS_ID", "GOID", "NOID", "NDOID", "NUOID", "LENGTH_KM", "LENGTH_DOWN_KM", "UPLAND_SKM", "RIV_ORD", "HYFALL", "DIS_AV_CMS", 
                                            "VOLUME_TCM", "INC", "CON_ID", "BAS_NAME", "ISO_NAME", "BB_NAME", "BB_ID", "BB_DIS_ORD", "BB_OCEAN", "BB_LEN_KM", "BB_VOL_TCM", "ERO_YLD_TON", "FLD"]

                    # create a list of fields in barrier feature class
                    stream_fields = arcpy.ListFields(streams.valueAsText)
                    # create list of field names from list of fields
                    stream_field_names = [f.name for f in stream_fields]

                    # iterate throught the list of required fields and check if it is in barrier feature class
                    for field in required_stream_fields:
                        if field not in stream_field_names:
                            streams.setErrorMessage("The required field '{}' was not found in the streams feature class. Check that the field exists and is spelled correctly.".format(field)) 

        ### check lakes fc for required fields
        if lakes.altered:
            if lakes.valueAsText != None:
                if arcpy.Exists(lakes.valueAsText):
                    # create list of required field names
                    required_lake_fields = [ "GOID", "GOOD", "Lake_type", "SED_ACC", "IN_STREAM", "IN_CATCH", "Vol_total", "Dis_avg", "Res_time"]

                    # create a list of fields in barrier feature class
                    lake_fields = arcpy.ListFields(lakes.valueAsText)
                    # create list of field names from list of fields
                    lake_field_names = [f.name for f in lake_fields]

                    # iterate throught the list of required fields and check if it is in barrier feature class
                    for field in required_lake_fields:
                        if field not in lake_field_names:
                            lakes.setErrorMessage("The required field '{}' was not found in the lakes feature class. Check that the field exists and is spelled correctly.".format(field)) 


        ##### check the minimum stream length threshold #####
        
        # minimum stream length threshold setting
        min_length_param =  parameters[8]
        min_length_val =  parameters[8].value


        # check if the 
        if min_length_val != None:
            if min_length_param.altered: # if any of the parameter have been altered
                
                # if the value is less than or equal to zero
                if min_length_val <=0:
                    # create an error message
                    min_length_param.setErrorMessage("Minimum stream length threshold must be greater than zero.")

        ##### check the upstream and downstream range factors #####

        # use dam level discharge factor setting
        use_dam_level_df_param =  parameters[10]
        use_dam_level_df_val =  parameters[10].valueAsText

        # static upstream discharge factor 
        upstream_drf_param = parameters[11]
        upstream_drf_val = parameters[11].value
        
        # static downstream discharge factor
        downstream_drf_param = parameters[12]
        downstream_drf_val = parameters[12].value
        
        # Upstream discharge factor field to use if "use dam level discharge factor" is set to YES
        DFU_field_param = parameters[13]
        DFU_field_val = parameters[13].valueAsText
        
        # Downstream discharge factorfield to use if "use dam level discharge factor" is set to YES
        DFD_field_param = parameters[14]
        DFD_field_val = parameters[14].valueAsText   


        
        # check if the use_dam_level_df_param has been altered
        # this tree will need to be checked if the values change or the yes/no option changes
        if use_dam_level_df_param.altered or upstream_drf_param.altered or  downstream_drf_param.altered: # if any of the parameter have been altered

            # Check static values    
            # these are required if "use dam level discharge factor" is set to NO
            # if use dam level df is set to no
            if use_dam_level_df_val == "NO": 
                
                # check if upstream drf is set to an appropriate value
                if upstream_drf_val == None: # if the upstream discharge factor parameter is empty, create an error
                    upstream_drf_param.setErrorMessage("If use dam level discharge factor is set to 'NO', an upstream discharge range factor must be specified.")

                
                if upstream_drf_val < 1:   # if the parameter is less than 1, create an error message.
                    upstream_drf_param.setErrorMessage("Upstream Discharge Range Factor must be greater than or equal to 1.")

                # check if downstream drf is set to an appropriate value
                if downstream_drf_param == None: # if the parameter isn't empty
                    downstream_drf_param.setErrorMessage("If use dam level discharge factor is set to 'NO', a downstream discharge range factor must be specified.")

                if downstream_drf_val < 1:   # if the parameter is less than 1, create error message
                    downstream_drf_param.setErrorMessage("Downstream Discharge Range Factor must be greater than or equal to 1.")

            
            # make fields required if "use dam level discharge factor" is set to YES
            # if use dam level df is set to no
            if use_dam_level_df_val == "YES": 
                # check if upstream drf field has been specifed
                if DFU_field_val == None: # if the upstream discharge field parameter is empty, create an error
                    DFU_field_param.setErrorMessage("If use dam level discharge factor is set to 'YES', a field with upstream discharge range factors in the dams feature class must be specified.")

                if DFD_field_val == None: # if the Downstream discharge field parameter is empty, create an error
                    DFD_field_param.setErrorMessage("If use dam level discharge factor is set to 'YES', a field with downstream discharge range factors in the dams feature class must be specified.")

        
        ##### check scenario table parameters  #####
        # assign parameter to a more descriptive variable name 
        scenario_param = parameters[19]
        scenarios_table = parameters[19].value

        ### check that the scenario name fits arcmaps field naming conventions ###
        # guidlines for field names: https://desktop.arcgis.com/en/arcmap/latest/manage-data/tables/fundamentals-of-adding-and-deleting-fields.htm#GUID-8E190093-8F8F-4132-AF4F-B0C9220F76B3
        # note: it is important that scenario name is checked first since it cannot be editted without deleting the whole row
        # therefore, validation checks for the name need to be at the bottom of the list so that they are checked last and will be the most recent error message that is displayed
        if scenarios_table != None:
            if scenario_param.altered:
                list_of_names = []
                
                for i in range(len(scenarios_table)):
                    scen_name = scenarios_table[i][0]

                    # check if filter_thresh is a valid percent value (0-100%) 
                    if scenarios_table[i][15] > 100 or scenarios_table[i][15] <= 0:
                        scenario_param.setErrorMessage("filter_thresh for scenario named: " + "'" + str(scenarios_table[i][0]) + "'" +
                        " is not a valid percent. Values for filter_thresh should be between 0 and 100.") 
 
                   # check if fld_damp is a valid percent value (0-100%) 
                    if scenarios_table[i][14] >100 or scenarios_table[i][14] <= 0:
                        scenario_param.setErrorMessage("fld_damp for scenario named: " + "'" + str(scenarios_table[i][0]) + "'" +
                        " is not a valid percent. Values for fld_damp should be between 0 and 100.") 
           
                    # check if CSI_threshold is a valid percent value (0-100%) 
                    if scenarios_table[i][13] > 100 or scenarios_table[i][13] <= 0:
                        scenario_param.setErrorMessage("CSI_threshold for scenario named: " + "'" + str(scenarios_table[i][0]) + "'" +
                        " is not a valid percent. Values for CSI_threshold should be greater than 0 and less than or equal to 100.")                                 
   
                    # check if scenario weight settings sum to 100. 
                    # for each row of the table sum the weight values
                    scenario_weigths_sum = sum([scenarios_table[i][7],     
                    scenarios_table[i][8], scenarios_table[i][9], scenarios_table[i][10],
                    scenarios_table[i][11], scenarios_table[i][12]]) 

                    # if the sum doesn't equal 100, create a warning
                    if scenario_weigths_sum != 100:      
                        scenario_param.setErrorMessage("Scenario weights for scenario named: " + "'" + str(scenarios_table[i][0]) + "'" +
                        " does not equal 100. The sum of the weights must equal 100%.")        

                    
                    ### check if indicator fields exist in the streams feature class ###
                    
                    # assign parameter to a more descriptive variable name 
                    streams_fc = parameters[3]
                    
                    # check if the streams_ fc parameter has been set
                    if streams_fc != None:
                        if streams_fc.altered:
                            if arcpy.Exists(streams_fc.valueAsText):
                                # create a list of the fields in the streams feature class
                                field_names = [f.name for f in arcpy.ListFields(streams_fc.value)] 
                                
                                # check if the field selected is in the 'streams feature class'
                                if scenarios_table[i][6] not in field_names:
                                    scenario_param.setErrorMessage("The current field selected for 'urb_indicator' does not exist in the 'Streams feature class'.\
                                    Please select an existing field from the 'Streams feature class'.")

                                # check if the field selected is in the 'streams feature class'
                                if scenarios_table[i][5] not in field_names:
                                    scenario_param.setErrorMessage("The current field selected for 'rdd_indicator' does not exist in the 'Streams feature class'.\
                                    Please select an existing field from the 'Streams feature class'.")

                                # check if the field selected is in the 'streams feature class'
                                if scenarios_table[i][4] not in field_names:
                                    scenario_param.setErrorMessage("The current field selected for 'use_indicator' does not exist in the 'Streams feature class'.\
                                    Please select an existing field from the 'Streams feature class'.")

                                # check if the field selected is in the 'streams feature class'
                                if scenarios_table[i][3] not in field_names:
                                    scenario_param.setErrorMessage("The current field selected for 'sed_indicator' does not exist in the 'Streams feature class'.\
                                    Please select an existing field from the 'Streams feature class'.")

                                # check if the field selected is in the 'streams feature class'
                                if scenarios_table[i][2] not in field_names:
                                    scenario_param.setErrorMessage("The current field selected for 'dor_indicator' does not exist in the 'Streams feature class'.\
                                    Please select an existing field from the 'Streams feature class'.")

                                # check if the field selected is in the 'streams feature class'
                                if scenarios_table[i][1] not in field_names:
                                    scenario_param.setErrorMessage("The current field selected for 'dof_indicator' does not exist in the 'Streams feature class'.\
                                    Please select an existing field from the 'Streams feature class'.")                                         
                            

                        

                    
                    ### check the scenario name ###
                    if scen_name not in list_of_names: 
                        list_of_names.append(scen_name)
                        characters = list(scen_name)

                        # check if the scenario name is under 10 characters
                        if len(characters)>9:
                            scenario_param.setErrorMessage("Scenario name" + "'" + str(scenarios_table[i][0]) + "'" + " is " + str(len(characters)) 
                            + " characters long. Scenario names must be under 10 characters in length.")

                        # check if the first character is a number
                        if characters[0].isdigit():
                            scenario_param.setErrorMessage("Scenario name" + "'" + str(scenarios_table[i][0]) + "'" + " starts with a number. " + 
                            "Scenario names cannot start with a number.")
                        
                        # check if the first character is an underscore
                        if characters[0] is "_":
                            scenario_param.setErrorMessage("Scenario name" + "'" + str(scenarios_table[i][0])
                            + "'" 
                            + " starts with an underscore ("+ "'"+"_" +"'"+")." 
                            + " Scenario names cannot start with an underscore ("+"'"+"_" +"'"+").")
                        
                        # check that the name has only alphanumeric characters or underscore
                        for character in characters:
                            if not character.isalnum() and not "_":
                                scenario_param.setErrorMessage("Scenario name" + "'" + str(scenarios_table[i][0]) + "'"
                                + " contains a character that is not alphanumeric or an underscore." 
                                + " Scenario names can only contain alphanumeric characters and underscores.")

                        
                    
                    else:
                        scenario_param.setErrorMessage("Scenario name" + "'" + str(scenarios_table[i][0]) + "'"
                        + " is used for more than one scenario. Each scenario must have a unique name.")
                        scenario_name_checked = True

        return


    def execute(self, parameters, messages):
        """The source code of the tool. Creates 
        results folder and Config.xls before executing the python code
        created by Guenther."""
    
        ##### create excel sheet from the input parameters #####

        # assign parameters to more descriptive variables
        run_options = parameters[0].valueAsText
        dams_fc = parameters[1].valueAsText
        bench_fc = parameters[2].valueAsText
        streams_fc = parameters[3].valueAsText
        lakes_fc = parameters[4].valueAsText
        output_folder_para = parameters[5].valueAsText
        barrier_inc_field = parameters[6].valueAsText
        update_mode = parameters[7].valueAsText
        min_length = parameters[8].value # for this one we want the value and not .valueAsText
        dof_field = parameters[9].valueAsText
        use_dam_level_df = parameters[10].valueAsText
        drf_upstream = parameters[11].value # for this one we want the value and not .valueAsText
        drf_downstream = parameters[12].value # for this one we want the value and not .valueAsText
        DFU_field = parameters[13].valueAsText
        DFD_field = parameters[14].valueAsText
        dof_mode = parameters[15].value # for this one we want the value and not .valueAsText
        dor_field = parameters[16].valueAsText
        svol_field = parameters[17].valueAsText
        sed_field = parameters[18].valueAsText
        scenarios_table = parameters[19].value # for this one we want the value and not .valueAsText

        # for path parameters, if they are a layer in the .mxd only the name is put in the excel
        # get the file path by describing the object
        # by using arcpy.Describe().name this also enables getting the path to the file if the name has been changed in 
        # the TOC
        dams_path = os.path.join(arcpy.Describe(dams_fc).path, arcpy.Describe(dams_fc).name)
        bench_path = os.path.join(arcpy.Describe(bench_fc).path, arcpy.Describe(bench_fc).name)
        streams_path = os.path.join(arcpy.Describe(streams_fc).path, arcpy.Describe(streams_fc).name)
        lakes_path = os.path.join(arcpy.Describe(lakes_fc).path, arcpy.Describe(lakes_fc).name)

        ##### create excel file using panadas (pd) #####

        # create time stamp
        var.time_stamp = tools.get_stamp()

        # Create a time-stamped folder under "output_folder"
        var.output_folder = os.path.join(output_folder_para, "Results_" + var.time_stamp)
        # create output file path
        var.config_excel = os.path.join(var.output_folder, "config_" + var.time_stamp + ".xls")

        # create output folder 
        tools.create_path(var.output_folder, messages)

        ### create start sheet entries ###
        # first, convert run options from string into YES/NO values
        # splits the run_options into a list
        run_options_list = run_options.split(";")
            
        # if DOF is selected,
        if "DOF" in run_options_list:
            # set value to yes
                run_dof = 'YES'
            
        # if DOF is not selected
        else: 
            run_dof = 'NO'

        # if DOR is selected,
        if "DOR" in run_options_list:
            # set value to yes
            run_dor = 'YES'
            
        # if DOR is not selected
        else: 
            run_dor = 'NO'

        # if SED is selected,
        if "SED" in run_options_list:
            # set value to yes
                run_sed = 'YES'

        # if SED is not selected
        else: 
            run_sed = 'NO'

        # if CSI is selected
        if "CSI" in run_options_list:
            # set value to yes
            run_csi = 'YES'
            
        # if CSI is not selected
        else: 
            run_csi = 'NO'   

        # create pandas data frame of the start sheet entries
        start_sheet = pd.DataFrame({'Key': ['settings_sheet', 'scenarios_sheet', '', 'run_dof', 'run_dor', 'run_sed', 'run_csi'],
                'Value': ['SETTINGS','SCENARIOS', '',run_dof, run_dor, run_sed, run_csi]
                })
        

        ##### create settings sheet entries #####

        settings_sheet = pd.DataFrame({ 'Category': ['path settings','path settings','path settings','path settings','path settings',
        'general settings','general settings','general settings',
        'DOF settings','DOF settings','DOF settings','DOF settings','DOF settings','DOF settings','DOF settings',
        'DOR settings', 'DOR settings',
        'SED settings'],
        'Key': ['dams_fc', 'bench_fc', 'streams_fc', 'lakes_fc', 'output_folder', 
        'barrier_inc_field', 'update_mode', 'min_length', 
        'dof_field', 'drf_upstream', 'drf_downstream', 'use_dam_level_df', 'DFU_field', 'DFD_field',  'dof_mode',
        'dor_field', 'svol_field',
        'sed_field'],
        'Value': [dams_path, bench_path, streams_path, lakes_path, output_folder_para,
        barrier_inc_field, update_mode, min_length,
        dof_field, drf_upstream, drf_downstream, use_dam_level_df, DFU_field, DFD_field, dof_mode, 
        dor_field, svol_field, 
        sed_field]
        })

        ##### create scenarios sheet entries #####
        scenarios_sheet = pd.DataFrame( scenarios_table, columns = [
                "scenario_name",
                "dof_indicator",
                "dor_indicator",
                "sed_indicator",
                "use_indicator",
                "rdd_indicator",
                "urb_indicator",
                "dof_weight",
                "dor_weight",
                "sed_weight",
                "use_weight",
                "rdd_weight",
                "urb_weight", 
                'csi_thresh', 
                'fld_damp',
                'filter_thresh', 
                'to_process', 
                'to_export'])


        # write each dataframe to the corresponding worksheet
        with pd.ExcelWriter(var.config_excel) as writer:
            start_sheet.to_excel(writer, sheet_name='START', index=False)
            settings_sheet.to_excel(writer, sheet_name = 'SETTINGS', index= False)
            scenarios_sheet.to_excel(writer, sheet_name = 'SCENARIOS', index= False)


        # ----- ## ----- ## ----- ## ----- ## ----- ## ----- ## ----- ## ----- ##

        if __name__ == '__main__':
            
            fra_start.main(messages)
        
        return
